﻿//-----------------------------------------------------------------------------------------
// File:  Assembler.cs
// Title: Propeller Assembler
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace ASMProp
{
    class Assembler
    {
        private const int NOOP = 0x12;

        private int address;
        private string[] saLine = new string[6];
        private Dictionary<string, int> symbols;
        private List<CodeLine> code_lines;
        private Dictionary<string, int> returns;
        private Equation parser;
        private bool first_res;
        private string header, footer;
        private string base_label;

        private int error_count;
        private int line_no;

        public int ErrorCount { get { return error_count; } }
        public bool ListFile { get; set; }
        public bool BinFile { get; set; }

        //-----------------------------------------------------------------------------------------
        public Assembler()
        {
            symbols = new Dictionary<string, int>();
            code_lines = new List<CodeLine>();
            returns = new Dictionary<string, int>();
            parser = new Equation(symbols);
            ListFile = false;
            BinFile = false;
        }

        //-----------------------------------------------------------------------------------------
        public void Assemble(string file_name)
        {
            try
            {
                line_no = 0;
                error_count = 0;
                first_res = false;
                base_label = "";

                InitSymbols();
                ReadSourceFile(file_name);
                FillDstSrc();

                if (BinFile)
                    WriteBinFile(Path.ChangeExtension(file_name, ".bin"));
                else
                    WriteHexFile(Path.ChangeExtension(file_name, ".hex"));

                if (ListFile) WriteListFile(file_name);
            }
            catch (Exception ex)
            {
                error_count++;
                Console.WriteLine(ex.Message);
            }
        }

        //-----------------------------------------------------------------------------------------
        private void WriteBinFile(string file_name)
        {
            using (BinaryWriter bw = new BinaryWriter(File.Open(file_name, FileMode.Create)))
            {
                foreach (CodeLine cl in code_lines)
                {
                    if (cl.type == CLType.CODE || cl.type == CLType.LONG)
                        bw.Write(cl.opcode);
                }
            }
        }

        //-----------------------------------------------------------------------------------------
        private void WriteHexFile(string file_name)
        {
            bool first = true;

            using (StreamWriter sw = new StreamWriter(file_name))
            {
                if (header != null) sw.WriteLine(header);

                foreach (CodeLine cl in code_lines)
                {
                    if (cl.type == CLType.CODE || cl.type == CLType.LONG)
                    {
                        if (first) first = false;
                        else
                        {
                            sw.Write(',');
                            if ((cl.address & 3) == 0) sw.WriteLine();
                        }

                        sw.Write(string.Format("0x{0,8:X8}", cl.opcode));
                    }
                }

                if (footer != null) sw.WriteLine(footer);
            }
        }

        //-----------------------------------------------------------------------------------------
        private void InitSymbols()
        {
            symbols.Add("TRUE", 1);
            symbols.Add("FALSE", 0);

            symbols.Add("PAR", 0x1F0);
            symbols.Add("CNT", 0x1F1);
            symbols.Add("INA", 0x1F2);
            symbols.Add("INB", 0x1F3);
            symbols.Add("OUTA", 0x1F4);
            symbols.Add("OUTB", 0x1F5);
            symbols.Add("DIRA", 0x1F6);
            symbols.Add("DIRB", 0x1F7);
            symbols.Add("CTRA", 0x1F8);
            symbols.Add("CTRB", 0x1F9);
            symbols.Add("FRQA", 0x1FA);
            symbols.Add("FRQB", 0x1FB);
            symbols.Add("PHSA", 0x1FC);
            symbols.Add("PHSB", 0x1FD);
            symbols.Add("VCFG", 0x1FE);
            symbols.Add("VSCL", 0x1FF);
        }

        //-----------------------------------------------------------------------------------------
        private int DecodeSrcDst(string s)
        {
            int offset = 0;
            int weight = 10;
            int value = 0;
            string alphabet;
            char c;

            if (returns.ContainsKey(s)) return returns[s];

            if (char.IsLetter(s[0]))
            {
                if (symbols.ContainsKey(s) == false)
                    throw new Exception("Symbol not declared: " + s);

                return symbols[s];
            }

            if (s[0] == '$') weight = 16;
            if (s[0] == '%') weight = 2;
            if (weight != 10) offset++;

            alphabet = "0123456789ABCDEF".Substring(0, weight);

            while (offset < s.Length)
            {
                c = char.ToUpper(s[offset]);
                if (c != '_')
                {
                    int digit = alphabet.IndexOf(c);
                    if (digit < 0)
                        throw new Exception("Invalid number: " + s);
                    value = (value * weight) + digit;
                }
                offset++;
            }

            if ((value & 0xFFFFFE00) != 0)
                throw new Exception("Value out of range: " + s);

            return value;
        }

        //-----------------------------------------------------------------------------------------
        private void FillDstSrc()
        {
            foreach (CodeLine cl in code_lines)
            {
                if (cl.error_msg != null) continue;

                try
                {
                    if (cl.type == CLType.CODE)
                    {
                        if (cl.dst_field != null)
                            cl.opcode |= DecodeSrcDst(cl.dst_field) << 9;

                        if (cl.src_field != null)
                            cl.opcode |= DecodeSrcDst(cl.src_field);
                    }
                    else if (cl.type == CLType.LONG)
                        cl.opcode = parser.ParseInt(cl.src_field);
                }
                catch (Exception ex)
                {
                    if (error_count < 10)
                        Console.WriteLine(ex.Message);
                    cl.error_msg = ex.Message;
                    error_count++;
                }
            }
        }

        //-----------------------------------------------------------------------------------------
        private void WriteListFile(string file_name)
        {
            StringBuilder sb = new StringBuilder();
            string list_file_name = Path.ChangeExtension(file_name, ".lst");
            string s, t, sline;
            StreamWriter list_file = null;
            StreamReader source_file = null;
            int index;
            CodeLine cl;

            try
            {
                list_file = new StreamWriter(list_file_name);
                source_file = new StreamReader(file_name);
                line_no = 0;
                index = 0;

                while (index < code_lines.Count && source_file.EndOfStream == false)
                {
                    sline = source_file.ReadLine();
                    line_no++;
                    cl = code_lines[index];

                    if (cl.source_line == line_no)
                    {
                        index++;
                        switch (cl.type)
                        {
                            case CLType.CODE:
                                t = string.Format("{0,2:X2} {1,1:X1} {2,1:X1} {3,3:X3} {4,3:X3}",
                                    (cl.opcode >> 26) & 0x3F, (cl.opcode>> 22) & 15, (cl.opcode>> 18) & 15,
                                    (cl.opcode >> 9) & 0x1FF, cl.opcode & 0x1FF);
                                s = string.Format("{0,3:X3} {1} {2}", cl.address, t, sline);
                                list_file.WriteLine(s);
                                break;

                            case CLType.LONG:
                                s = string.Format("{0,3:X3} {1,8:X8}       {2}", cl.address, cl.opcode, sline);
                                list_file.WriteLine(s);
                                break;

                            case CLType.RES:
                                s = string.Format("{0,3:X3}                {1}", cl.address, sline);
                                list_file.WriteLine(s);
                                break;

                            case CLType.ORG:
                                s = string.Format("{0,3:X3}                {1}", cl.address, sline);
                                list_file.WriteLine(s);
                                break;

                            default:
                                list_file.WriteLine("                   " + sline);
                                break;
                        }

                        if (cl.error_msg != null)
                            list_file.WriteLine("ERROR >>> " + cl.error_msg);
                    }
                    else
                    {
                        list_file.WriteLine("                   " + sline);
                    }
                }

                while (source_file.EndOfStream == false)
                {
                    sline = source_file.ReadLine();
                    list_file.WriteLine("                   " + sline);
                }

            }
            finally
            {
                if (list_file != null) list_file.Close();
                if (source_file != null) source_file.Close();
            }
        }

        //-----------------------------------------------------------------------------------------
        private void ReadSourceFile(string file_name)
        {
            using (StreamReader source_file = new StreamReader(file_name))
            {
                while (source_file.EndOfStream == false)
                {
                    string sline = source_file.ReadLine();
                    line_no++;

                    if (sline.Length > 0)
                    {
                        if (sline.StartsWith("#header") || sline.StartsWith("' #header"))
                            header = ProcessDirective(sline);
                        else if (sline.StartsWith("#footer") || sline.StartsWith("' #footer"))
                            footer = ProcessDirective(sline);
                        else
                            ProcessLine(sline);
                    }
                }
            }
        }

        //-----------------------------------------------------------------------------------------
        private string ProcessDirective(string line)
        {
            string[] sa = line.Split('\"');
            if (sa.Length < 3)
            {
                error_count++;
                Console.WriteLine("Invalid directive: " + line);
            }

            return sa[1];
        }

        //-----------------------------------------------------------------------------------------
        private void ProcessLine(string sLine)
        {
            CodeLine cl = new CodeLine();
            cl.source_line = line_no;
            cl.opcode = 0;
            cl.address = address;
            cl.type = CLType.COMMENT;

            try
            {
                ParseLine(sLine);

                if (saLine[0] != "")
                {
                    if (symbols.ContainsKey(saLine[0]))
                        throw new Exception("Duplicate label : " + saLine[0]);

                    symbols.Add(saLine[0], address);
                }

                if (saLine[2] == "") return;
                if (first_res && saLine[2] != "RES" && saLine[2] != "FIT")
                    throw new Exception("RES must be last");

                switch (saLine[2])
                {
                    case "ORG": Org(ref cl); break;
                    case "LONG": LONG(ref cl); break;
                    case "RES": RES(ref cl); break;

                    case "ABS": OpDstSrc(0x2A2, ref cl); break;
                    case "ABSNEG": OpDstSrc(0x2B2, ref cl); break;
                    case "ADD": OpDstSrc(0x202, ref cl); break;
                    case "ADDABS": OpDstSrc(0x222, ref cl); break;
                    case "ADDS": OpDstSrc(0x342, ref cl); break;
                    case "ADDSX": OpDstSrc(0x362, ref cl); break;
                    case "ADDX": OpDstSrc(0x322, ref cl); break;
                    case "AND": OpDstSrc(0x182, ref cl); break;
                    case "ANDN": OpDstSrc(0x192, ref cl); break;
                    case "CMP": OpDstSrc(0x210, ref cl); break;
                    case "CMPS": OpDstSrc(0x300, ref cl); break;
                    case "CMPSUB": OpDstSrc(0x382, ref cl); break;
                    case "CMPSX": OpDstSrc(0x310, ref cl); break;
                    case "CMPX": OpDstSrc(0x330, ref cl); break;
                    case "DJNZ": OpDstSrc(0x392, ref cl); break;
                    case "HUBOP": OpDstSrc(0x030, ref cl); break;
                    case "JMPRET": OpDstSrc(0x172, ref cl); break;
                    case "MAX": OpDstSrc(0x132, ref cl); break;
                    case "MAXS": OpDstSrc(0x112, ref cl); break;
                    case "MIN": OpDstSrc(0x122, ref cl); break;
                    case "MINS": OpDstSrc(0x102, ref cl); break;
                    case "MOV": OpDstSrc(0x282, ref cl); break;
                    case "MOVD": OpDstSrc(0x152, ref cl); break;
                    case "MOVI": OpDstSrc(0x162, ref cl); break;
                    case "MOVS": OpDstSrc(0x142, ref cl); break;
                    case "MUXC": OpDstSrc(0x1C2, ref cl); break;
                    case "MUXNC": OpDstSrc(0x1D2, ref cl); break;
                    case "MUXNZ": OpDstSrc(0x1F2, ref cl); break;
                    case "MUXZ": OpDstSrc(0x1E2, ref cl); break;
                    case "NEG": OpDstSrc(0x292, ref cl); break;
                    case "NEGC": OpDstSrc(0x2C2, ref cl); break;
                    case "NEGNC": OpDstSrc(0x2D2, ref cl); break;
                    case "NEGNZ": OpDstSrc(0x2F2, ref cl); break;
                    case "NEGZ": OpDstSrc(0x2E2, ref cl); break;
                    case "OR": OpDstSrc(0x1A2, ref cl); break;
                    case "RCL": OpDstSrc(0x0D2, ref cl); break;
                    case "RCR": OpDstSrc(0x0C2, ref cl); break;
                    case "RDBYTE": OpDstSrc(0x002, ref cl); break;
                    case "RDLONG": OpDstSrc(0x022, ref cl); break;
                    case "RDWORD": OpDstSrc(0x012, ref cl); break;
                    case "REV": OpDstSrc(0x0F2, ref cl); break;
                    case "ROL": OpDstSrc(0x092, ref cl); break;
                    case "ROR": OpDstSrc(0x082, ref cl); break;
                    case "SAR": OpDstSrc(0x0E2, ref cl); break;
                    case "SHL": OpDstSrc(0x0B2, ref cl); break;
                    case "SHR": OpDstSrc(0x0A2, ref cl); break;
                    case "SUB": OpDstSrc(0x212, ref cl); break;
                    case "SUBABS": OpDstSrc(0x232, ref cl); break;
                    case "SUBS": OpDstSrc(0x352, ref cl); break;
                    case "SUBSX": OpDstSrc(0x372, ref cl); break;
                    case "SUBX": OpDstSrc(0x332, ref cl); break;
                    case "SUMC": OpDstSrc(0x242, ref cl); break;
                    case "SUMNC": OpDstSrc(0x252, ref cl); break;
                    case "SUMNZ": OpDstSrc(0x272, ref cl); break;
                    case "SUMZ": OpDstSrc(0x262, ref cl); break;
                    case "TEST": OpDstSrc(0x180, ref cl); break;
                    case "TESTN": OpDstSrc(0x190, ref cl); break;
                    case "TJNZ": OpDstSrc(0x3A0, ref cl); break;
                    case "TJZ": OpDstSrc(0x3B0, ref cl); break;
                    case "WAITCNT": OpDstSrc(0x3E2, ref cl); break;
                    case "WAITPEQ": OpDstSrc(0x3C0, ref cl); break;
                    case "WAITPNE": OpDstSrc(0x3D0, ref cl); break;
                    case "WAITVID": OpDstSrc(0x3F0, ref cl); break;
                    case "WRBYTE": OpDstSrc(0x000, ref cl); break;
                    case "WRLONG": OpDstSrc(0x020, ref cl); break;
                    case "WRWORD": OpDstSrc(0x010, ref cl); break;
                    case "XOR": OpDstSrc(0x1B2, ref cl); break;

                    case "NOP": Op(0x0000, ref cl); break;

                    case "CALL": OpCall(ref cl); break;
                    case "JMP": OpJmp(ref cl); break;
                    case "RET": OpRet(ref cl); break;

                    case "CLKSET": OpDst(0x031, 0, ref cl); break;
                    case "COGID": OpDst(0x033, 1, ref cl); break;
                    case "COGINIT": OpDst(0x031, 2, ref cl); break;
                    case "COGSTOP": OpDst(0x031, 3, ref cl); break;
                    case "LOCKCLR": OpDst(0x031, 7, ref cl); break;
                    case "LOCKNEW": OpDst(0x031, 4,ref cl); break;
                    case "LOCKRET": OpDst(0x031, 5, ref cl); break;
                    case "LOCKSET": OpDst(0x031, 6, ref cl); break;

                    case "FIT":
                        if (address > 511)
                            throw new Exception("Code does not fit");
                        break;

                    default:
                        throw new Exception("Invalid nemonic : " + saLine[1]);
                }

                AddWriteFlags(ref cl);
                AddConditionCode(ref cl);
            }
            catch (Exception ex)
            {
                cl.error_msg = ex.Message;
                if (error_count < 10) Console.WriteLine(ex.Message);
                error_count++;
            }

            code_lines.Add(cl);
        }

        //-----------------------------------------------------------------------------------------
        private void AddWriteFlags(ref CodeLine cl)
        {
            string[] sa = saLine[5].Split(',');

            foreach (string s in sa)
            {
                switch (s.Trim())
                {
                    case "": break;
                    case "WZ": cl.opcode |= 0x02000000; break;
                    case "WC": cl.opcode |= 0x01000000; break;
                    case "WR": cl.opcode |= 0x00800000; break;
                    case "NR": cl.opcode &= ~0x00800000; break;
                    default:
                        throw new Exception("Invalid Write FLag");
                }
            }
        }

        //-----------------------------------------------------------------------------------------
        private void AddConditionCode(ref CodeLine cl)
        {
            switch (saLine[1])
            {
                case "":
                case "IF_ALWAYS":
                    cl.opcode |= 0x003C0000; break;
                case "IF_NEVER":
                    cl.opcode |= 0x00000000; break;
                case "IF_E":
                case "IF_Z":
                    cl.opcode |= 0x00280000; break;
                case "IF_NE":
                case "IF_NZ":
                    cl.opcode |= 0x00140000; break;
                case "IF_C":
                case "IF_B":
                    cl.opcode |= 0x00300000; break;
                case "IF_AE":
                case "IF_NC":
                    cl.opcode |= 0x000C0000; break;
                case "IF_C_EQ_Z":
                case "IF_Z_EQ_C":
                    cl.opcode |= 0x00240000; break;
                case "IF_C_NE_Z":
                case "IF_Z_NE_C":
                    cl.opcode |= 0x00180000; break;
                case "IF_C_AND_Z":
                case "IF_Z_AND_C":
                    cl.opcode |= 0x00200000; break;
                case "IF_C_AND_NZ":
                case "IF_NZ_AND_C":
                    cl.opcode |= 0x00100000; break;
                case "IF_NC_AND_Z":
                case "IF_Z_AND_NC":
                    cl.opcode |= 0x00080000; break;
                case "IF_A":
                case "IF_NC_AND_NZ":
                case "IF_NZ_AND_NC":
                    cl.opcode |= 0x00040000; break;
                case "IF_BE":
                case "IF_C_OR_Z":
                case "IF_Z_OR_C":
                    cl.opcode |= 0x00380000; break;
                case "IF_C_OR_NZ":
                case "IF_NZ_OR_C":
                    cl.opcode |= 0x00340000; break;
                case "IF_NC_OR_Z":
                case "IF_Z_OR_NC":
                    cl.opcode |= 0x002C0000; break;
                case "IF_NC_OR_NZ":
                case "IF_NZ_OR_NC":
                    cl.opcode |= 0x001C0000; break;
                default:
                    throw new Exception("Invalid Write FLag");
            }
        }

        //-----------------------------------------------------------------------------------------
        private void OpCall(ref CodeLine cl)
        {
            cl.opcode = 0x5CC00000;
            cl.type = CLType.CODE;

            if (saLine[3] == null || saLine[3][0] != '#')
                throw new Exception("Call label missing #");

            if (saLine[3].StartsWith("#:"))
                cl.src_field = base_label + saLine[3].Substring(1);
            else
                cl.src_field = saLine[3].Substring(1);

            cl.dst_field = saLine[3].Substring(1) + "_RET";

            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void OpJmp(ref CodeLine cl)
        {
            cl.opcode = 0x5C000000;
            cl.type = CLType.CODE;

            if (saLine[3].StartsWith("#"))
            {
                cl.opcode |= 0x00400000;
                if (saLine[3].StartsWith("#:"))
                    cl.src_field = base_label + saLine[3].Substring(1);
                else
                    cl.src_field = saLine[3].Substring(1);
            }
            else
                cl.src_field = saLine[3];

            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void OpRet(ref CodeLine cl)
        {
            if (saLine[0] == "")
                throw new Exception("Missing return statement label");
            cl.opcode = 0x5C400000;
            cl.type = CLType.CODE;
            returns.Add(saLine[0], address);
            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void Op(int op, ref CodeLine cl)
        {
            cl.opcode = op << 18;
            cl.type = CLType.CODE;
            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void OpSrc(int op, ref CodeLine cl)
        {
            cl.opcode = op << 22;
            cl.type = CLType.CODE;

            if (saLine[3].Length == 0)
                throw new Exception("Missing parameter");

            DecodeSrc(ref cl);
            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void OpDst(int op, int src, ref CodeLine cl)
        {
            cl.opcode = (op << 22) | src;
            cl.type = CLType.CODE;

            if (saLine[3].Length == 0)
                throw new Exception("Missing parameter");

            cl.dst_field = saLine[3];
            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void OpDstSrc(int op, ref CodeLine cl)
        {
            cl.opcode = op << 22;
            cl.type = CLType.CODE;
            cl.dst_field = saLine[3];

            if (saLine[3].Length == 0 || saLine[4].Length == 0)
                throw new Exception("Missing parameter");

            DecodeSrc(ref cl);
            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void DecodeSrc(ref CodeLine cl)
        {
            if (saLine[4].StartsWith("#:"))
            {
                cl.opcode |= 0x00400000;
                cl.src_field = base_label + saLine[4].Substring(1);
            }
            else if (saLine[4].StartsWith("#"))
            {
                cl.opcode |= 0x00400000;
                cl.src_field = saLine[4].Substring(1);
            }
            else
            {
                cl.src_field = saLine[4];
            }
        }

        //-----------------------------------------------------------------------------------------
        private void Org(ref CodeLine cl)
        {
            if (saLine[3] == "")
                throw new Exception("Missing operands");

            address = parser.ParseInt(saLine[3]);
            cl.address = address;
            cl.type = CLType.ORG;
        }

        //-----------------------------------------------------------------------------------------
        private void LONG(ref CodeLine cl)
        {
            if (saLine[3] == "")
                throw new Exception("Missing value");

            if (saLine[4] != "")
                throw new Exception("Too many values");

            cl.type = CLType.LONG;
            cl.src_field = saLine[3];
            address++;
        }

        //-----------------------------------------------------------------------------------------
        private void RES(ref CodeLine cl)
        {
            cl.type = CLType.RES;
            int count = parser.ParseInt(saLine[3]);
            address += count;
            first_res = true;
        }

        //-----------------------------------------------------------------------------------------
        // saLine[0] -> Label
        // saLine[1] -> IF_ Condition
        // saLine[2] -> Mnemonic
        // saLine[3] -> Parm1
        // saLine[4] -> Parm2
        // saLine[5] -> Write Flags
        //-----------------------------------------------------------------------------------------
        private void ParseLine(string s)
        {
            int i, j;
            string t;

            // Init strings
            for (i = 0; i < 6; i++) saLine[i] = "";

            // Ignore blank lines
            if (s.Length == 0) return;

            // Find first field
            i = j = 0;
            ParseField(ref s, ref i, ref j);
            if (i >= s.Length || s[i] == '\'') return;

            // If first character is in first column
            if (i == 0)
            {
                // Save the label
                if (s[0] == ':')
                {
                    saLine[0] = base_label + s.Substring(0, j).ToUpper();
                }
                else
                {
                    saLine[0] = s.Substring(0, j).ToUpper();
                    base_label = saLine[0];
                }

                // Find next field
                ParseField(ref s, ref i, ref j);
                if (i >= s.Length || s[i] == '\'') return;
            }

            // Check for IF_ Conditions
            t = s.Substring(i, j - i).ToUpper();
            if (t.StartsWith("IF_"))
            {
                // Save for IF_ Conditions
                saLine[1] = t;

                // Find next field
                ParseField(ref s, ref i, ref j);
                if (i >= s.Length || s[i] == '\'') return;
            }

            // Save the Mnemonic
            saLine[2] = s.Substring(i, j - i).ToUpper();

            // Find parm1 field
            ParseField(ref s, ref i, ref j);
            if (i >= s.Length || s[i] == '\'') return;

            // Save the parm1
            saLine[3] = s.Substring(i, j - i).ToUpper();
            if (saLine[3].StartsWith("#:"))
                saLine[3] = string.Format("#{0}{1}", base_label, saLine[3].Substring(1));

            // Find next field
            ParseField(ref s, ref i, ref j);
            if (i >= s.Length || s[i] == '\'') return;

            // Check for second parm
            if (s[i] == ',')
            {
                // Find second parm field
                ParseField(ref s, ref i, ref j);
                if (i >= s.Length || s[i] == '\'') return;

                // Save the parm2
                saLine[4] = s.Substring(i, j - i).ToUpper();

                // Find write flag field
                ParseField(ref s, ref i, ref j);
                if (i >= s.Length || s[i] == '\'') return;
            }

            while (i < s.Length && s[i] != '\'')
            {
                // Save the write flags
                saLine[5] += s.Substring(i, j - i).ToUpper() + ' ';

                // Find write flag field
                ParseField(ref s, ref i, ref j);
            }
        }

        //-----------------------------------------------------------------------------------------
        private void ParseField(ref string s, ref int i, ref int j)
        {
            i = j;

            // Find next non-space character
            while (i < s.Length && char.IsWhiteSpace(s[i])) i++;
            j = i + 1;

            if (i >= s.Length) return;
            if (s[i] == '\'') return;
            if (s[i] == ',') return;

            // Find next space character
            while (j < s.Length)
            {
                if (char.IsWhiteSpace(s[j])) return;
                if (s[j] == '\'') return;
                if (s[j] == ',') return;
                j++;                
            }
        }
    }
}
