﻿//-----------------------------------------------------------------------------------------
// File:  CodeGen.cs
// Title: Asm Code Generator
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace PCC
{
    class CodeGen
    {
        private OpcodeTable ot;
        private SymbolTable st;
        private FuncTable ft;
        private string errmsg;
        private string stats;

        private StreamReader sr;
        private StreamWriter sw;
        private OpCodeEntry opcode;
        private OpCodeEntry.ComparisonT comparison;
        private int opcode_count, init_symbol, res_symbol, util_count;
        private bool IMul, SMul, RMul, UDiv, SDiv, RDiv, UMod, SMod, SinCos, ASinCos, Tan, Log, Exp, Sqrt;
        private VarType type;

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public CodeGen(OpcodeTable _ot, SymbolTable _st, FuncTable _ft)
        {
            ot = _ot;
            st = _st;
            ft = _ft;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public string ErrorMsg { get { return errmsg; } }
        public string Stats { get { return stats; } }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public void Run(string input_filename, string output_filename)
        {
            try
            {
                opcode_count = 0;
                init_symbol = res_symbol = util_count = 0;

                sr = new StreamReader(input_filename);
                sw = new StreamWriter(output_filename, false, Encoding.Unicode);

                IMul = SMul = RMul = UDiv = SDiv = RDiv = UMod = SMod = SinCos = ASinCos = Tan = Log = Exp = Sqrt = false;
                errmsg = "";

                DumpOpcodes();
                DumpUtilities();
                DumpSymbolTable();
                emit1p("fit", "");

                stats = "Opcodes = " + opcode_count.ToString();
                stats += "   Utilities = " + util_count.ToString();
                stats += "   Symbols = " + init_symbol.ToString();
                stats += "   Reserves = " + res_symbol.ToString();
                stats += "   Total = " + (opcode_count + init_symbol + res_symbol + util_count).ToString();
            }
            catch (Exception ex)
            {
                errmsg = ex.Message + "  At line " + opcode.LineNo.ToString();
            }
            finally
            {
                sr.Close();
                sw.Close();
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void DumpOpcodes()
        {
            int op_count = ot.PC;
            int line_count = 0;
            int i;
            string s;
            bool blank = false;
            bool need_org = true;

            for (i = 0; i < op_count; i++)
            {
                opcode = ot[i];

                while (line_count < opcode.LineNo)
                {
                    if (blank)
                    {
                        sw.WriteLine();
                        blank = false;
                    }
                    if (sr.EndOfStream == true) break;
                    s = sr.ReadLine();
                    sw.WriteLine("\' " + s);
                    line_count++;
                }

                if (need_org)
                {
                    need_org = false;
                    sw.WriteLine("                        org     0");
                }

                blank = true;
                type = GetType(opcode.Dst);
                if (opcode.Comparison != OpCodeEntry.ComparisonT.NONE)
                    comparison = opcode.Comparison;

                switch (opcode.OpCode)
                {
                    case OpCodes.MOV: emit2p("MOV"); break;
                    case OpCodes.MOVI: emit2p("MOVI"); break;
                    case OpCodes.MOVS: emit2p("MOVS"); break;
                    case OpCodes.MOVD: emit2p("MOVD"); break;

                    case OpCodes.CMP_EQ:
                    case OpCodes.CMP_NE:
                    case OpCodes.CMP_LT:
                    case OpCodes.CMP_GT:
                    case OpCodes.CMP_LE:
                    case OpCodes.CMP_GE:
                        if (opcode.Type == VarType.INT || opcode.Type == VarType.REAL)
                            emit2p("CMPS");
                        else
                            emit2p("CMP");
                        break;

                    case OpCodes.CMP_LC:
                        emit1p("NOP", "");
                        break;
                    case OpCodes.CMP_LS:
                        emit1p("NOP", "");
                        break;

                    case OpCodes.ADD:
                        if (opcode.Type != VarType.UINT)
                            emit2p("ADDS");
                        else
                            emit2p("ADD");
                        break;

                    case OpCodes.SUB:
                        if (opcode.Type != VarType.UINT)
                            emit2p("SUBS");
                        else
                            emit2p("SUB");
                        break;

                    case OpCodes.SUBX:
                        emit2p("SUBX");
                        break;

                    case OpCodes.INCR:
                        if (opcode.Type != VarType.UINT)
                            emit2p("ADDS", "#1");
                        else
                            emit2p("ADD", "#1");
                        break;

                    case OpCodes.DECR:
                        if (opcode.Type != VarType.UINT)
                            emit2p("SUBS", "#1");
                        else
                            emit2p("SUB", "#1");
                        break;

                    case OpCodes.IMUL:
                        math_call("IMultiply");
                        break;

                    case OpCodes.RMUL:
                        math_call("RMultiply");
                        break;

                    case OpCodes.IMOD:
                    case OpCodes.IDIV:
                        if (opcode.Type != VarType.UINT)
                            math_call("SDivide");
                        else
                            math_call("UDivide");
                        break;

                    case OpCodes.RDIV:
                        math_call("RDivide");
                        break;

                    case OpCodes.SIN: math_call("Sin"); break;
                    case OpCodes.COS: math_call("Cos"); break;
                    case OpCodes.TAN: math_call("Tan"); break;
                    case OpCodes.ASIN: math_call("ASin"); break;
                    case OpCodes.ACOS: math_call("ACos"); break;

                    case OpCodes.SQRT:
                        math_call("Sqrt");
                        emit1p("SHL", "math_r1, #8");
                        break;

                    case OpCodes.ISQRT: math_call("Sqrt"); break;
                    case OpCodes.LOG: math_call("ULog"); break;
                    case OpCodes.EXP: math_call("UExp"); break;

                    case OpCodes.WR_BYTE: emit2p("WRBYTE"); break;
                    case OpCodes.WR_WORD: emit2p("WRWORD"); break;
                    case OpCodes.WR_LONG: emit2p("WRLONG"); break;

                    case OpCodes.RD_BYTE: emit2p("RDBYTE"); break;
                    case OpCodes.RD_WORD: emit2p("RDWORD"); break;
                    case OpCodes.RD_LONG: emit2p("RDLONG"); break;

                    case OpCodes.BW_TEST: emit2p("TEST"); break;
                    case OpCodes.BW_OR: emit2p("OR"); break;
                    case OpCodes.BW_AND: emit2p("AND"); break;
                    case OpCodes.BW_XOR: emit2p("XOR"); break;
                    case OpCodes.BW_AND_NOT: emit2p("ANDN"); break;
                    case OpCodes.BW_NOT:
                        opcode.Src = st.Find("math_ones", "");
                        st.MarkUsed(opcode.Src.Value);
                        emit2p("XOR", "math_ones");
                        break;

                    case OpCodes.LABEL: sw.WriteLine(opcode.Name); break;
                    case OpCodes.CALL: emit1p("CALL", '#' + opcode.Name); break;
                    case OpCodes.JMP: emit1p("JMP", '#' + opcode.Name); break;
                    case OpCodes.JMP_F: jmp_f(); break;
                    case OpCodes.JMP_T: jmp_t(); break;
                    case OpCodes.JMP_DZ: emit2p("DJNZ", '#' + opcode.Name); break;

                    case OpCodes.RET:
                        sw.WriteLine("{0,-23} RET", opcode.Name);
                        sw.WriteLine("'---------------------------------------------------------");
                        opcode_count++;
                        break;

                    case OpCodes.SHIFT_LEFT: emit2p("SHL"); break;
                    case OpCodes.SHIFT_RIGHT:
                        if (opcode.Type != VarType.UINT)
                            emit2p("SAR");
                        else
                            emit2p("SHR");
                        break;

                    case OpCodes.ROTATE_LEFT: emit2p("ROL"); break;
                    case OpCodes.ROTATE_RIGHT: emit2p("ROR"); break;

                    case OpCodes.REV: emit2p("REV"); break;
                    case OpCodes.CMPSUB: emit2p("CMPSUB"); break;

                    case OpCodes.NEG: emit2p("NEG"); break;
                    case OpCodes.ABS: emit2p("ABS"); break;
                    case OpCodes.ABSN: emit2p("ABSNEG"); break;

                    case OpCodes.RTOI:
                        if (opcode.Type != VarType.UINT)
                            emit2p("SAR", "#16");
                        else
                            emit2p("SHR", "#16");
                        break;
                    case OpCodes.ITOR: emit2p("SHL", "#16"); break;

                    case OpCodes.CEIL:
                        emit2p("ADD", "math_real_mask");
                        emit2p("ANDN", "math_real_mask");
                        st.MarkUsed("math_real_mask", "");
                        break;
                    case OpCodes.FLOOR:
                        emit2p("ANDN", "math_real_mask");
                        st.MarkUsed("math_real_mask", "");
                        break;
                    case OpCodes.TRUNC:
                        emit2p("ABS", "");
                        emit2p("ANDN", "math_real_mask");
                        emit2p("NEGC", "");
                        st.MarkUsed("math_real_mask", "");
                        break;
                    case OpCodes.FRACT:
                        emit2p("ABS", "");
                        emit2p("AND", "math_real_mask");
                        emit2p("NEGC", "");
                        st.MarkUsed("math_real_mask", "");
                        break;
                    case OpCodes.ROUND:
                        emit2p("ADD", "math_half");
                        emit2p("ANDN", "math_real_mask");
                        st.MarkUsed("math_half", "");
                        st.MarkUsed("math_real_mask", "");
                        break;

                    case OpCodes.WAITCNT: emit2p("WAITCNT"); break;
                    case OpCodes.WAITPEQ: emit2p("WAITPEQ"); break;
                    case OpCodes.WAITPNE: emit2p("WAITPNE"); break;
                    case OpCodes.WAITVID: emit2p("WAITVID"); break;

                    case OpCodes.LOCKNEW: emit1p("LOCKNEW", opcode.Dst); break;
                    case OpCodes.LOCKSET: emit1p("LOCKSET", opcode.Dst); break;
                    case OpCodes.LOCKCLR: emit1p("LOCKCLR", opcode.Dst); break;
                    case OpCodes.LOCKRET: emit1p("LOCKRET", opcode.Dst); break;

                    case OpCodes.LIMIT_MIN:
                        if (opcode.Type != VarType.UINT) emit2p("MINS");
                        else emit2p("MIN");
                        break;
                    case OpCodes.LIMIT_MAX:
                        if (opcode.Type != VarType.UINT) emit2p("MAXS");
                        else emit2p("MAX");
                        break;

                    case OpCodes.RESET: reset_array(); break;

                    case OpCodes.NOP: emit1p("NOP", ""); break;
                    case OpCodes.NULL: break;
                    case OpCodes.ENDEX: break;

                    default:
                        throw new Exception("Opcode not yet implimented: " + opcode.OpCode.ToString());
                }
            }

            while (sr.EndOfStream == false)
            {
                s = sr.ReadLine();
                if (s.StartsWith("/#"))
                    sw.WriteLine(s.Substring(2));
                else
                    sw.WriteLine("\' " + s);
                line_count++;
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void reset_array()
        {
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void DumpUtilities()
        {
            if (IMul || RMul) WriteUtil(UMulAsm);
            if (RMul) WriteUtil(RMulAsm);
            if (UDiv || SDiv || RDiv || UMod || SMod || Tan) WriteUtil(UDivAsm);
            if (SDiv || RDiv || SMod || Tan) WriteUtil(SDivAsm);
            if (RDiv || Tan) WriteUtil(RDivAsm);
            if (SinCos) WriteUtil(SinCosAsm);
            if (Tan) WriteUtil(TanAsm);
            if (ASinCos) WriteUtil(ASinCosAsm);
            if (Log) WriteUtil(LogAsm);
            if (Exp) WriteUtil(ExpAsm);
            if (Sqrt) WriteUtil(SqrtAsm);

            if (IMul || SMul || RMul || UDiv || SDiv || RDiv || UMod || SMod || SinCos || Tan || ASinCos || Log || Exp || Sqrt)
            {
                st.MarkUsed("math_t", "");
                st.MarkUsed("math_s", "");
                st.MarkUsed("math_p1", "");
                st.MarkUsed("math_p2", "");
                st.MarkUsed("math_r1", "");
                st.MarkUsed("math_r2", "");
                st.MarkUsed("math_ones", "");
            }

            if (SinCos || Tan || ASinCos)
            {
                st.MarkUsed("math_90", "");
                st.MarkUsed("math_180", "");
            }
        }

        private void WriteUtil(string[] sa)
        {
            foreach (string s in sa)
            {
                util_count++;
                sw.WriteLine(s);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void DumpSymbolTable()
        {
            int i;
            Symbol sym;

            for (i = 0; i < st.Count; i++)
            {
                sym = st[i];
                if (sym.Used == true && (sym.State == SymState.INITIALIZED || sym.State == SymState.CONST))
                {
                    sw.WriteLine("{0,-23} LONG    {1}", sym.Name, sym.Value.ToString());
                    init_symbol++;
                }
            }

            for (i = 0; i < st.Count; i++)
            {
                sym = st[i];
                if (sym.Used == true && sym.State == SymState.VOID)
                {
                    sw.WriteLine("{0,-23} RES     1", sym.Name);
                    res_symbol++;
                }
            }

            sw.WriteLine("'---------------------------------------------------------");
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void jmp_t()
        {
            string s = "";

            switch (comparison)
            {
                case OpCodeEntry.ComparisonT.EQ: s = "IF_Z"; break;
                case OpCodeEntry.ComparisonT.NE: s = "IF_NZ"; break;
                case OpCodeEntry.ComparisonT.LT: s = "IF_C"; break;
                case OpCodeEntry.ComparisonT.LE: s = "IF_C_OR_Z"; break;
                case OpCodeEntry.ComparisonT.GT: s = "IF_NC_AND_NZ"; break;
                case OpCodeEntry.ComparisonT.GE: s = "IF_NC"; break;
                case OpCodeEntry.ComparisonT.LC: s = "IF_NC"; break;
                case OpCodeEntry.ComparisonT.LS: s = "IF_C"; break;
            }

            sw.WriteLine("   {0,-20} JMP     #{1}", s, opcode.Name);
            opcode_count++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void jmp_f()
        {
            string s = "";

            switch (comparison)
            {
                case OpCodeEntry.ComparisonT.EQ: s = "IF_NZ"; break;
                case OpCodeEntry.ComparisonT.NE: s = "IF_Z"; break;
                case OpCodeEntry.ComparisonT.LT: s = "IF_NC"; break;
                case OpCodeEntry.ComparisonT.LE: s = "IF_NC_AND_NZ"; break;
                case OpCodeEntry.ComparisonT.GT: s = "IF_C_OR_Z"; break;
                case OpCodeEntry.ComparisonT.GE: s = "IF_C"; break;
                case OpCodeEntry.ComparisonT.LC: s = "IF_C"; break;
                case OpCodeEntry.ComparisonT.LS: s = "IF_NC"; break;
            }

            sw.WriteLine("   {0,-20} JMP     #{1}", s, opcode.Name);
            opcode_count++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void math_call(string inst)
        {
            switch (inst)
            {
                case "IMultiply": IMul = true; break;
                case "SMultiply": SMul = true; break;
                case "RMultiply": RMul = true; break;
                case "UDivide": UMod = true; break;
                case "SDivide": SMod = true; break;
                case "RDivide": RDiv = true; break;
                case "Sin":
                case "Cos": SinCos = true; break;
                case "Tan": Tan = true; break;
                case "ASin":
                case "ACos": ASinCos = true; break;
                case "ULog":
                case "RLog": Log = true; break;
                case "UExp":
                case "RExp": Exp = true; break;
                case "Sqrt": Sqrt = true; break;
            }

            sw.WriteLine("                        {0,-7} #{1}", "CALL", inst);
            opcode_count++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private string FlagField()
        {
            string ffld = "";
            
            switch (opcode.Comparison)
            {
                case OpCodeEntry.ComparisonT.EQ:
                case OpCodeEntry.ComparisonT.NE:
                    ffld = "WZ";
                    break;
                case OpCodeEntry.ComparisonT.LC:
                case OpCodeEntry.ComparisonT.LS:
                    ffld = "WC";
                    break;
                case OpCodeEntry.ComparisonT.GT:
                case OpCodeEntry.ComparisonT.LT:
                case OpCodeEntry.ComparisonT.GE:
                case OpCodeEntry.ComparisonT.LE:
                    ffld = "WZ, WC";
                    break;
            }

            return ffld;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void emit1p(string inst, string dfld)
        {
            string ffld = FlagField();
            sw.WriteLine("                        {0,-7} {1} {2}", inst, dfld, ffld);
            opcode_count++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void emit1p(string inst, ValType op1)
        {
            string ffld = FlagField();
            st.MarkUsed(op1.Value);
            string dfld = st[op1.Value].Name;
            sw.WriteLine("                        {0,-7} {1} {2}", inst, dfld, ffld);
            opcode_count++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void emit2p(string inst, string p2 = null)
        {
            string ffld = FlagField();
            string sfld, dfld;

            ValType dst = opcode.Dst;
            ValType src = opcode.Src;

            while (dst.Type == VarType.OPCODE)
                dst = ot[dst.Value].Dst;
            dfld = st[dst.Value].Name;
            st.MarkUsed(dst.Value);

            if (p2 == "")
                sfld = dfld;
            else if (p2 != null)
                sfld = p2;
            else
            {
                while (src.Type == VarType.OPCODE)
                    src = ot[src.Value].Dst;
                sfld = st[src.Value].Name;

                Symbol ssym = st[src.Value];

                if ((ssym.Type == VarType.AINT || ssym.Type == VarType.INT || ssym.Type == VarType.UINT) && 
                    ssym.State == SymState.CONST &&
                    ssym.Value >= 0 && 
                    ssym.Value < 512)
                    sfld = '#' + ssym.Value.ToString();
                else
                    st.MarkUsed(src.Value);
            }

            if (dfld == "ina")
                throw new Exception("Attemp to write to INA");

            sw.WriteLine("                        {0,-7} {1}, {2}  {3}", inst, dfld, sfld, ffld);
            opcode_count++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private VarType GetType(ValType op1)
        {
            while (op1.Type == VarType.OPCODE)
                op1 = ot[op1.Value].Dst;

            return op1.Type;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        string[] UMulAsm = 
        {
            "IMultiply               mov     math_r1, #0",
            "                        mov     math_r2, #0",
            "                        mov     math_t, #0",
            ":IM0                    shr     math_p1, #1  wz, wc",
            "        if_z_and_nc     jmp     #IMultiply_ret",
            "        if_nc           jmp     #:IM1",
            "                        add     math_r1, math_p2  wc",
            "                        addx    math_r2, math_t",
            ":IM1                    shl     math_p2, #1  wc",
            "                        rcl     math_t, #1",
            "                        jmp     #:IM0",
            "IMultiply_ret           ret",
            "'---------------------------------------------------------"
        };

        string[] RMulAsm = 
        {
            "RMultiply               call    #IMultiply",
            "                        shr     math_r1, #16",
            "                        shl     math_r2, #16",
            "                        or      math_r1, math_r2",
            "RMultiply_ret           ret",
            "'---------------------------------------------------------"
        };

        string[] UDivAsm = 
        {
            "UDivide                 mov     math_r1, #0",
            "                        mov     math_r2, #0",
            "                        mov     math_t, #32",
            ":UD0                    shl     math_p1, #1  wc",
            "                        rcl     math_r2, #1",
            "                        cmpsub  math_r2, math_p2  wc",
            "                        rcl     math_r1, #1",
            "                        djnz    math_t, #:UD0",
            "UDivide_ret             ret",
            "'---------------------------------------------------------"
        };

        string[] SDivAsm = 
        {
            "SDivide                 mov     math_s, #0",
            "                        cmps    math_p1, #0  wc",
            "        if_c            neg     math_p1, math_p1",
            "        if_c            mov     math_s, #3",
            "                        cmps    math_p2, #0  wc",
            "        if_c            neg     math_p2, math_p2",
            "        if_c            xor     math_s, #1",
            "                        call    #UDivide",
            "                        shr     math_s, #1  wc",
            "        if_c            neg     math_r1, math_r1",
            "                        shr     math_s, #1  wc",
            "        if_c            neg     math_r2, math_r2",
            "SDivide_ret             ret",
            "'---------------------------------------------------------"
        };

        string[] RDivAsm = 
        {
            "RDivide                 mov     math_s, #0",
            "                        cmps    math_p1, #0  wc",
            "        if_c            neg     math_p1, math_p1",
            "        if_c            mov     math_s, #1",
            "                        cmps    math_p2, #0  wc",
            "        if_c            neg     math_p2, math_p2",
            "        if_c            xor     math_s, #1",
            "                        mov     math_r1, #0",
            "                        mov     math_r2, #0",
            "                        mov     math_t, #48",
            ":RD0                    shl     math_p1, #1  wc",
            "                        rcl     math_r2, #1",
            "                        cmpsub  math_r2, math_p2  wc",
            "                        rcl     math_r1, #1",
            "                        djnz    math_t, #:RD0",
            "                        rcr     math_s, math_s  wc",
            "                        negc    math_r1, math_r1",
            "RDivide_ret             ret",
            "'---------------------------------------------------------"
        };

        // 90Deg = 128, 180Deg = 256, 360Deg = 512
        string[] SinCosAsm = 
        {
            "Cos                     add     math_p1, math_90",
            "Sin                     test    math_p1, math_90  wc",
            "                        test    math_p1, math_180  wz",
            "                        negc    math_p1, math_p1",
            "                        shr     math_p1, #11",
            "                        cmp     math_90, math_p1  wc",
            "        if_c            andn    math_p1, math_90",
            "                        or      math_p1, SinCos_table",
            "                        rdword  math_r1, math_p1",
            "                        negnz   math_r1, math_r1",
            "Sin_ret",
            "Cos_ret                 ret",
            "SinCos_table            long    $E000",
            "'---------------------------------------------------------"
        };
    
        string[] TanAsm = 
        {
            "Tan                     mov     math_p2, math_p1",
            "                        call    #Cos",
            "                        mov     math_p1, math_p2",
            "                        mov     math_p2, math_r1",
            "                        call    #Sin",
            "                        mov     math_p1, math_r1",
            "                        call    #RDivide",
            "Tan_ret                 ret",
            "'---------------------------------------------------------"
        };

        string[] ASinCosAsm = 
        {
            "ASin                    cmps    math_p1, #0  wc",
            "                        negc    math_p1, math_p1",
            "                        rcl     math_p2, #1",
            "                        mov     math_r2, math_p1",
            "                        shr     math_r2, #16  wz",
            "        if_z            call    #BinSrch",
            "        if_nz           mov     math_r1, math_90",
            "                        ror     math_p2, #1  wc",
            "                        negc    math_r1, math_r1",
            "ASin_ret                ret",
            "'---------------------------------------------------------",
            "ACos                    call    #ASin",
            "                        subs    math_r1, math_90",
            "                        neg     math_r1, math_r1",
            "ACos_ret                ret",
            "'---------------------------------------------------------",
            "BinSrch                 mov     math_r1, BinSrch_table",
            "                        mov     math_s, #1",
            "                        shl     math_s, #11",
            "                        mov     math_t, #0",
            ":BS0                    or      math_r1, math_s",
            "                        rdword  math_t, math_r1",
            "                        cmps    math_p1, math_t  wc",
            "        if_c            andn    math_r1, math_s",
            "                        shr     math_s, #1  wz",
            "        if_nz           jmp     #:BS0",
            "                        andn    math_r1, BinSrch_table",
            "                        shl     math_r1, #11",
            "BinSrch_ret             ret",
            "BinSrch_table           long    $E001",
            "'---------------------------------------------------------"
        };

        string[] LogAsm = 
        {
            "ULog                    mov     math_t, #31",
            ":L1                     shl     math_p1, #1  wc",
            "        if_nc           sub     math_t, #1",
            "        if_nc           jmp     #:L1",
            "                        shr     math_p1, #20",
            "                        or      math_p1, math_log_table",
            "                        rdword  math_r1, math_p1",
            "                        shl     math_t, #16",
            "                        or      math_r1, math_t",
            "ULog_ret                ret",
            "math_log_table          long    $C000",
            "'---------------------------------------------------------"
        };

        string[] ExpAsm = 
        {
            "UExp                    mov     math_r1, math_p1",
            "                        shr     math_r1, #4",
            "                        and     math_r1, math_exp_mask",
            "                        or      math_r1, math_exp_table",
            "                        rdword  math_r1, math_r1",
            "                        shl     math_r1, #15",
            "                        or      math_r1, math_msb",
            "                        shr     math_p1, #16",
            "                        xor     math_p1, #31",
            "                        shr     math_r1, math_p1",
            "UExp_ret                ret",
            "math_exp_mask           long    $0FFE",
            "math_exp_table          long    $D000",
            "math_msb                long    $80000000",
            "'---------------------------------------------------------"
        };

        string[] SqrtAsm = 
        {
            "Sqrt                    mov     math_s,#0",
            "                        mov     math_r1,#0",
            "                        mov     math_t,#16",
            ":loop                   shl     math_p1,#1 wc",
            "                        rcl     math_s,#1",
            "                        shl     math_p1,#1 wc",
            "                        rcl     math_s,#1",
            "                        shl     math_r1,#2",
            "                        or      math_r1,#1",
            "                        cmpsub  math_s,math_r1 wc",
            "                        shr     math_r1,#2",
            "                        rcl     math_r1,#1",
            "                        djnz    math_t,#:loop",
            "Sqrt_ret                ret",
            "'---------------------------------------------------------"
        };
    }
}
