﻿//-----------------------------------------------------------------------------------------
// File:  Equation.cs
// Title: Equation Parser
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------
// <expression> ::= <additive-expression>
//                | <shift-expression> '<<' <additive-expression>
//                | <shift-expression> '>>' <additive-expression> ;
//
// <additive-expression> ::= <multiplicative-expression>
//                         | <additive-expression> '+' <multiplicative-expression>
//                         | <additive-expression> '-' <multiplicative-expression> ;
//
// <multiplicative-expression> ::= <unary-expression>
//                               | <multiplicative-expression> '*' <unary-expression>
//                               | <multiplicative-expression> '/' <unary-expression>
//                               | <multiplicative-expression> '%' <unary-expression> ;
//
// <unary-expression> ::= <primary-expression>
//                      | '-' <primary-expression>
//                      | '+' <primary-expression>
//                      | '~' <primary-expression> ;
//
// <primary-expression> ::= <integer-constant>
//                        | <char-constant>
//                        | '(' <expression> ')' ;
//
// <integer-constant> ::= [0-9]+
//                      | '$' [0-9a-fA-F]+
//                      | '%' [01]+ ;
//-------------------------------------------------------------------------------------------------

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ASMProp
{
    public class Equation
    {
        private Dictionary<string, int> symbols;
        private enum Tokens { NONE, VALUE, PLUS, MINUS, BW_NOT, MULTIPLY, DIVIDE, MODULUS,
                              SHIFT_RIGHT, SHIFT_LEFT, PAREN_LEFT, PAREN_RIGHT };
        private List<EquElement> ElementList = new List<EquElement>();
        private int token_index, token_value;
        private Tokens token;

        //-------------------------------------------------------------------------------------------------
        public Equation(Dictionary<string, int> _symbols)
        {
            symbols = _symbols;
        }

        //-------------------------------------------------------------------------------------------------
        public int ParseInt(string equation_string)
        {
            string equation = FixSyntax(equation_string);

            ElementList.Clear();
            ParseString(ref equation);

            token_index = -1;
            GetToken();

            int op1 = expression();
            ElementList.Clear();
            return op1;
        }

        //-------------------------------------------------------------------------------------------------
        private string FixSyntax(string s)
        {
            s = ' ' + s + ' ';
            s = s.Replace('\t', ' ');
            return s;
        }

        //-------------------------------------------------------------------------------------------------
        // <expression> ::= <additive-expression>
        //                | <shift-expression> '<<' <additive-expression>
        //                | <shift-expression> '>>' <additive-expression> ;
        //-------------------------------------------------------------------------------------------------
        private int expression()
        {
            int op1 = additive_expression();

            while (token == Tokens.SHIFT_LEFT || token == Tokens.SHIFT_RIGHT)
            {
                Tokens tk = token;
                GetToken(); // '<<' '>>'
                switch(tk)
                {
                    case Tokens.SHIFT_LEFT:
                        op1 = op1 << additive_expression();
                        break;
                    case Tokens.SHIFT_RIGHT:
                        op1 = op1 >> additive_expression();
                        break;
                }
            }

            return op1;
        }

        //-------------------------------------------------------------------------------------------------
        // <additive-expression> ::= <multiplicative-expression>
        //                         | <additive-expression> '+' <multiplicative-expression>
        //                         | <additive-expression> '-' <multiplicative-expression> ;
        //-------------------------------------------------------------------------------------------------
        private int additive_expression()
        {
            int op1 = multiplicative_expression();

            while (token == Tokens.PLUS || token == Tokens.MINUS)
            {
                Tokens tk = token;
                GetToken(); // '+' '-'
                switch (tk)
                {
                    case Tokens.PLUS:
                        op1 = op1 + multiplicative_expression();
                        break;
                    case Tokens.MINUS:
                        op1 = op1 - multiplicative_expression();
                        break;
                }
            }

            return op1;
        }

        //-------------------------------------------------------------------------------------------------
        // <multiplicative-expression> ::= <unary-expression>
        //                               | <multiplicative-expression> '*' <unary-expression>
        //                               | <multiplicative-expression> '/' <unary-expression>
        //                               | <multiplicative-expression> '%' <unary-expression> ;
        //-------------------------------------------------------------------------------------------------
        private int multiplicative_expression()
        {
            int op1 = unary_expression();

            while (token == Tokens.MULTIPLY || token == Tokens.DIVIDE || token == Tokens.MODULUS)
            {
                Tokens tk = token;
                GetToken(); // '*' '/' '%'
                switch (tk)
                {
                    case Tokens.MULTIPLY:
                        op1 = op1 * unary_expression();
                        break;
                    case Tokens.DIVIDE:
                        op1 = op1 / unary_expression();
                        break;
                    case Tokens.MODULUS:
                        op1 = op1 % unary_expression();
                        break;
                }
            }

            return op1;
        }

        //-------------------------------------------------------------------------------------------------
        // <unary-expression> ::= <primary-expression>
        //                      | '-' <primary-expression>
        //                      | '+' <primary-expression>
        //                      | '~' <primary-expression> ;
        //-------------------------------------------------------------------------------------------------
        private int unary_expression()
        {
            switch (token)
            {
                case Tokens.BW_NOT:
                    GetToken(); // '~'
                    return ~primary_expression();
                case Tokens.MINUS:
                    GetToken(); // '-'
                    return -primary_expression();
                case Tokens.PLUS:
                    GetToken(); // '+'
                    return primary_expression();
                default:
                    return primary_expression();
            }
        }

        //-------------------------------------------------------------------------------------------------
        // <primary-expression> ::= <integer-constant>
        //                        | <char-constant>
        //                        | '(' <expression> ')' ;
        //-------------------------------------------------------------------------------------------------
        private int primary_expression()
        {
            int op1;
            switch (token)
            {
                case Tokens.VALUE:
                    op1 = token_value;
                    GetToken(); // VALUE
                    break;
                case Tokens.PAREN_LEFT:
                    GetToken(); // '('
                    op1 = expression();
                    if (token != Tokens.PAREN_RIGHT)
                        throw new Exception("Missing right paren");
                    GetToken(); // ')'
                    break;
                default:
                    throw new Exception("Invalid expression");
            }
            return op1;
        }

        //-------------------------------------------------------------------------------------------------
        private void GetToken()
        {
            ++token_index;

            if (token_index > ElementList.Count)
                throw new Exception("Invalid expression");

            if (token_index == ElementList.Count)
            {
                token = Tokens.NONE;
                token_value = 0;
            }
            else
            {
                token = ElementList[token_index].token;
                token_value = ElementList[token_index].value;
            }
        }

        //-------------------------------------------------------------------------------------------------
        private void ParseString(ref string equation_string)
        {
            char c;
            int value, index = 0, idx;
            const string hex_digits = "0123456789ABCDEF";
            StringBuilder sb = new StringBuilder();

            if (equation_string.Length == 0)
                throw new Exception("NULL equation string");

            while (index < equation_string.Length)
            {
                c = equation_string[index++];
                if (IsLabel(c, false))
                {
                    while (true)
                    {
                        sb.Append(c);
                        if (index >= equation_string.Length) break;

                        c = equation_string[index];
                        if (IsLabel(c, true) == false) break;
                        index++;
                    }
                    if (symbols.ContainsKey(sb.ToString()) == false)
                        throw new Exception("Symbol not defined: " + sb.ToString());
                    value = (int)symbols[sb.ToString()];
                    sb.Clear();
                    ElementList.Add(new EquElement(Tokens.VALUE, value));
                }
                else if (Char.IsDigit(c))
                {
                    while (true)
                    {
                        sb.Append(c);
                        if (index >= equation_string.Length) break;
                        c = equation_string[index];
                        if (Char.IsDigit(c) == false) break;
                        index++;
                    }
                    value = int.Parse(sb.ToString());
                    sb.Clear();
                    ElementList.Add(new EquElement(Tokens.VALUE, value));
                }
                else if (c == '$')
                {
                    value = 0;
                    while (true)
                    {
                        if (index >= equation_string.Length) break;
                        c = equation_string[index];
                        if (c != '_')
                        {
                            idx = hex_digits.IndexOf(c);
                            if (idx < 0) break;
                            value <<= 4;
                            value |= idx;
                        }
                        index++;
                    }
                    ElementList.Add(new EquElement(Tokens.VALUE, value));
                }
                else if (c == '%')
                {
                    value = 0;
                    while (true)
                    {
                        if (index >= equation_string.Length) break;
                        c = equation_string[index];
                        if (c != '0' && c != '1' && c != '_') break;
                        if (c != '_') value <<= 1;
                        if (c == '1') value |= 1;
                        index++;
                    }
                    ElementList.Add(new EquElement(Tokens.VALUE, value));
                }
                else
                {
                    switch (c)
                    {
                        case '+':
                            ElementList.Add(new EquElement(Tokens.PLUS, 0));
                            break;
                        case '-':
                            ElementList.Add(new EquElement(Tokens.MINUS, 0));
                            break;
                        case '~':
                            ElementList.Add(new EquElement(Tokens.BW_NOT, 0));
                            break;
                        case '*':
                            ElementList.Add(new EquElement(Tokens.MULTIPLY, 0));
                            break;
                        case '/':
                            ElementList.Add(new EquElement(Tokens.DIVIDE, 0));
                            break;
                        case '%':
                            ElementList.Add(new EquElement(Tokens.MODULUS, 0));
                            break;
                        case '(':
                            ElementList.Add(new EquElement(Tokens.PAREN_LEFT, 0));
                            break;
                        case ')':
                            ElementList.Add(new EquElement(Tokens.PAREN_RIGHT, 0));
                            break;
                        case '<':
                            if (equation_string[index] == '<')
                            {
                                ElementList.Add(new EquElement(Tokens.SHIFT_LEFT, 0));
                                index++;
                            }
                            else
                                throw new Exception("Invalid shift operator");
                            break;
                        case '>':
                            if (equation_string[index] == '>')
                            {
                                ElementList.Add(new EquElement(Tokens.SHIFT_RIGHT, 0));
                                index++;
                            }
                            else
                                throw new Exception("Invalid shift operator");
                            break;
                        case ' ':
                            break;
                        default:
                            throw new Exception("Invalid expression");
                    }
                }
            }
        }

        //-------------------------------------------------------------------------------------------------
        private bool IsLabel(char c, bool include_numbers)
        {
            if (char.IsLetter(c)) return true;
            if (include_numbers && char.IsNumber(c)) return true;
            if (c == '_' || c == '.' || c == '@') return true;
            return false;
        }

        //-------------------------------------------------------------------------------------------------
        private struct EquElement
        {
            public Tokens token;
            public int value;
            public EquElement(Tokens t, int v)
            {
                token = t;
                value = v;
            }
            public override string ToString()
            {
                if (token != Tokens.VALUE) return token.ToString();
                return token.ToString() + '=' + value.ToString();
            }
        }
    }
}
