﻿//-----------------------------------------------------------------------------------------
// File:  LexAn.cs
// Title: Lexical Analyzer
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

//-------------------------------------------------------------------------------------------------------------------------------------------------
namespace PCC
{
    //---------------------------------------------------------------------------------------------------------------------------------------------
    class LexAn
    {
        private int offset;
        private Tokens assign_op = Tokens.NONE;
        private int lineno;
        private int const_value;
        private string label;
        private VarType vartype;
        private string code = "";
        private BIFType biftype;
        private List<Token> token_list = new List<Token>();
        private Dictionary<string, string> defines = new Dictionary<string, string>();
        private StreamWriter temp_file;

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public LexAn()
        {
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public void Scan(string file_name, string temp_name)
        {
            lineno = 0;
            label = "";
            const_value = 0;
            vartype = VarType.NONE;
            token_list.Clear();

            using (temp_file = new StreamWriter(temp_name))
                ReadFile(file_name);

            offset = 0;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public void Next()
        {
            offset++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public void Next(Tokens current)
        {
            if (token_list[offset].token != current)
                throw new Exception("Expected: " + current.ToString());

            offset++;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public Tokens Peek
        {
            get
            {
                if (offset >= token_list.Count)
                    return Tokens.NONE;
                else
                    return token_list[offset].token;
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public int LineNo 
        { 
            get 
            {
                if (offset < token_list.Count)
                    return token_list[offset].source_line;
                else
                    return 1;
            }
        }

        public int Value { get { return token_list[offset].const_value; } }
        public string Label { get { return token_list[offset].label; } }
        public VarType Var_Type { get { return token_list[offset].var_type; } }
        public Tokens AssignOp { get { return token_list[offset].assign_op; } }
        public BIFType BIF_Type { get { return token_list[offset].bif_type; } }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        public int State
        {
            get { return offset; }
            set { offset = value; }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void ReadFile(string file_name)
        {
            using (StreamReader sr = new StreamReader(file_name))
            {
                while (sr.EndOfStream == false)
                {
                    code = sr.ReadLine();
                    temp_file.WriteLine(code);
                    lineno++;
                    code += " // ";

                    if (code.StartsWith("#include"))
                    {
                        string[] sa = code.Split('\"');
                        if (sa.Length < 3)
                            Console.WriteLine("Invalid #include: " + code);
                        else
                            ReadFile(sa[1]);
                    }
                    else if (code.StartsWith("#header") || code.StartsWith("#footer")) { }
                    else if (code.StartsWith("#define"))
                        ProcessDefine();
                    else
                        ProcessLine();
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void ProcessLine()
        {
            Tokens token;
            offset = 0;

            while (offset < code.Length)
            {
                // Skip white space
                while (offset < code.Length && char.IsWhiteSpace(code[offset]))
                    offset++;

                // Terminate on comments
                if (offset < code.Length - 1 && code[offset] == '/' && code[offset + 1] == '/')
                    return;

                if (Char.IsLetter(code[offset]))
                {
                    token = CheckLabel();
                    switch (token)
                    {
                        case Tokens.LABEL: token_list.Add(new Token(token, lineno, label)); break;
                        case Tokens.BIOP: token_list.Add(new Token(token, lineno, biftype)); break;
                        case Tokens.BIF: token_list.Add(new Token(token, lineno, biftype)); break;
                        case Tokens.VAR_TYPE: token_list.Add(new Token(token, lineno, "", vartype, 0)); break;
                        case Tokens.NONE: break;
                        default: token_list.Add(new Token(token, lineno)); break;
                    }
                }

                else if (Char.IsNumber(code[offset]))
                {
                    token = CheckConst();
                    token_list.Add(new Token(token, lineno, label, vartype, const_value));
                }

                else
                {
                    token = CheckSymbol();
                    token_list.Add(new Token(token, lineno, assign_op));
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void ProcessDefine()
        {
            int j, i = 7;
            string define_label, define_text;

            while (char.IsWhiteSpace(code[i]) == true) i++;
            j = i;
            while (char.IsWhiteSpace(code[j]) == false) j++;
            define_label = code.Substring(i, j - i).ToLower();

            if (defines.ContainsKey(define_label))
                throw new Exception("Duplicate #define label: " + define_label);

            define_text = code.Substring(j);
            i = define_text.IndexOf("//");
            if (i > -1) define_text = define_text.Substring(0, i);
            define_text = define_text.Trim();

            defines.Add(define_label, define_text);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private Tokens CheckLabel()
        {
            int len = 1;
            int start = offset;

            while (offset < code.Length)
            {
                char c = code[offset + len];
                if (c != '_' && Char.IsLetterOrDigit(c) == false) break;
                len++;
            }

            label = code.Substring(offset, len).ToLower();
            offset += len;

            switch (label)
            {
                case "if": return Tokens.IF;
                case "do": return Tokens.DO;
                case "for": return Tokens.FOR;
                case "nop": return Tokens.NOP;
                case "else": return Tokens.ELSE;
                case "case": return Tokens.CASE;
                case "break": return Tokens.BREAK;
                case "while": return Tokens.WHILE;
                case "return": return Tokens.RETURN;
                case "switch": return Tokens.SWITCH;
                case "default": return Tokens.DEFAULT;
                case "continue": return Tokens.CONTINUE;
                case "parity": return Tokens.PARITY;

                case "gbyte": return Tokens.GBYTE;
                case "gword": return Tokens.GWORD;
                case "glong": return Tokens.GLONG;

                case "sin": biftype = BIFType.SIN; return Tokens.BIF;
                case "cos": biftype = BIFType.COS; return Tokens.BIF;
                case "tan": biftype = BIFType.TAN; return Tokens.BIF;
                case "asin": biftype = BIFType.ASIN; return Tokens.BIF;
                case "acos": biftype = BIFType.ACOS; return Tokens.BIF;
                case "atan": biftype = BIFType.ATAN; return Tokens.BIF;
                case "log": biftype = BIFType.LOG; return Tokens.BIF;
                case "exp": biftype = BIFType.EXP; return Tokens.BIF;
                case "sqrt": biftype = BIFType.SQRT; return Tokens.BIF;
                case "isqrt": biftype = BIFType.ISQRT; return Tokens.BIF;

                case "rev": biftype = BIFType.REV; return Tokens.BIOP;
                case "abs": biftype = BIFType.ABS; return Tokens.BIOP;
                case "absn": biftype = BIFType.ABSN; return Tokens.BIOP;
                case "itor": biftype = BIFType.ITOR; return Tokens.BIOP;
                case "rtoi": biftype = BIFType.RTOI; return Tokens.BIOP;
                case "ceil": biftype = BIFType.CEIL; return Tokens.BIOP;
                case "floor": biftype = BIFType.FLOOR; return Tokens.BIOP;
                case "trunc": biftype = BIFType.TRUNC; return Tokens.BIOP;
                case "round": biftype = BIFType.ROUND; return Tokens.BIOP;
                case "fract": biftype = BIFType.FRACT; return Tokens.BIOP;
                
                case "clkset": biftype = BIFType.CLKSET; return Tokens.BIOP;
                case "cmpsub": biftype = BIFType.CMPSUB; return Tokens.BIOP;

                case "cogid": biftype = BIFType.COGID; return Tokens.BIOP;
                case "coginit": biftype = BIFType.COGINIT; return Tokens.BIOP;
                case "cogstop": biftype = BIFType.COGSTOP;  return Tokens.BIOP;

                case "int": vartype = VarType.INT; return Tokens.VAR_TYPE;
                case "uint": vartype = VarType.UINT; return Tokens.VAR_TYPE;
                case "real": vartype = VarType.REAL; return Tokens.VAR_TYPE;
                case "void": vartype = VarType.VOID; return Tokens.VAR_TYPE;

                case "true": label = "CONST_tru"; return Tokens.LABEL;
                case "false": label = "CONST_fls"; return Tokens.LABEL;
                case "odd": label = "CONST_odd"; return Tokens.LABEL;
                case "even": label = "CONST_even"; return Tokens.LABEL;

                case "lockset": return Tokens.LOCKSET;
                case "lockclr": return Tokens.LOCKCLR;
                case "locknew": return Tokens.LOCKNEW;
                case "lockret": return Tokens.LOCKRET;
                case "waitcnt": return Tokens.WAITCNT;
                case "waitpeq": return Tokens.WAITPEQ;
                case "waitpne": return Tokens.WAITPNE;
                case "waitvid": return Tokens.WAITVID;
            }

            if (defines.ContainsKey(label))
            {
                code = defines[label] + code.Substring(offset);
                offset = 0;
                return Tokens.NONE;
            }

            return Tokens.LABEL;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private Tokens CheckConst()
        {
            int start_offset = offset;
            int weight = 10;
            string alphabet;
            char c = ' ';

            string s = code.Substring(offset, 2);
            if (String.Compare(s, "0x", true) == 0) weight = 16;
            if (String.Compare(s, "0o", true) == 0) weight = 8;
            if (String.Compare(s, "0b", true) == 0) weight = 2;
            if (weight != 10) offset += 2;

            alphabet = "0123456789abcdef".Substring(0, weight);
            const_value = 0;

            while (offset < code.Length)
            {
                c = code[offset];
                c = char.ToLower(c);
                if (c == '.') break;
                if (c != '_')
                {
                    int digit = alphabet.IndexOf(c);
                    if (digit < 0) break;
                    const_value = (const_value * weight) + digit;
                }
                offset++;
            }

            if (c == '.')
            {
                offset++;
                while (offset < code.Length && char.IsDigit(code[offset])) 
                    offset++;

                label = code.Substring(start_offset, offset - start_offset);
                const_value = MakeReal(decimal.Parse(label));
                label = label.Replace('.', '_');
                vartype = VarType.REAL;
                return Tokens.REAL_CONSTANT;
            }

            label = code.Substring(start_offset, offset - start_offset);
            vartype = VarType.UINT;
            return Tokens.CONSTANT;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private Tokens CheckSymbol()
        {
            assign_op = Tokens.NONE;

            switch (code.Substring(offset, 3))
            {
                case "<<=": offset += 3; assign_op = Tokens.SHIFT_LEFT; return Tokens.ASSIGN;
                case ">>=": offset += 3; assign_op = Tokens.SHIFT_RIGHT; return Tokens.ASSIGN;
                case "<-=": offset += 3; assign_op = Tokens.ROTATE_LEFT; return Tokens.ASSIGN;
                case "->=": offset += 3; assign_op = Tokens.ROTATE_RIGHT; return Tokens.ASSIGN;
                case "&~=": offset += 3; assign_op = Tokens.BITWISE_AND_NOT; return Tokens.ASSIGN;
                case "#>=": offset += 3; assign_op = Tokens.LIMIT_MIN; return Tokens.ASSIGN;
                case "<#=": offset += 3; assign_op = Tokens.LIMIT_MAX; return Tokens.ASSIGN;
            }

            switch (code.Substring(offset, 2))
            {
                case "==": offset += 2; return Tokens.CMP_EQ;
                case "!=": offset += 2; return Tokens.CMP_NE;
                case ">=": offset += 2; return Tokens.CMP_GE;
                case "<=": offset += 2; return Tokens.CMP_LE;
                case "&&": offset += 2; return Tokens.LOGICAL_AND;
                case "||": offset += 2; return Tokens.LOGICAL_OR;
                case "^^": offset += 2; return Tokens.LOGICAL_XOR;
                case ">>": offset += 2; return Tokens.SHIFT_RIGHT;
                case "<<": offset += 2; return Tokens.SHIFT_LEFT;
                case "->": offset += 2; return Tokens.ROTATE_RIGHT;
                case "<-": offset += 2; return Tokens.ROTATE_LEFT;
                case "#>": offset += 2; return Tokens.LIMIT_MIN;
                case "<#": offset += 2; return Tokens.LIMIT_MAX;
                case "&~": offset += 2; return Tokens.BITWISE_AND_NOT;

                case "+=": offset += 2; assign_op = Tokens.PLUS; return Tokens.ASSIGN;
                case "-=": offset += 2; assign_op = Tokens.MINUS; return Tokens.ASSIGN;
                case "*=": offset += 2; assign_op = Tokens.MULTIPLY; return Tokens.ASSIGN;
                case "/=": offset += 2; assign_op = Tokens.DIVIDE; return Tokens.ASSIGN;
                case "%=": offset += 2; assign_op = Tokens.MODULO; return Tokens.ASSIGN;
                case "|=": offset += 2; assign_op = Tokens.BITWISE_OR; return Tokens.ASSIGN;
                case "&=": offset += 2; assign_op = Tokens.BITWISE_AND; return Tokens.ASSIGN;
                case "^=": offset += 2; assign_op = Tokens.BITWISE_XOR; return Tokens.ASSIGN;
            }

            switch (code.Substring(offset, 2).ToLower())
            {
                case ".i": offset += 2; return Tokens.DOTI;
                case ".d": offset += 2; return Tokens.DOTD;
                case ".s": offset += 2; return Tokens.DOTS;
            }

            switch (code[offset])
            {
                case '=': offset++; assign_op = Tokens.ASSIGN; return Tokens.ASSIGN;

                case ',': offset++; return Tokens.COMMA;
                case ';': offset++; return Tokens.SEMICOLON;
                case ':': offset++; return Tokens.COLON;
                case '(': offset++; return Tokens.PAREN_LEFT;
                case ')': offset++; return Tokens.PAREN_RIGHT;
                case '{': offset++; return Tokens.BRACE_LEFT;
                case '}': offset++; return Tokens.BRACE_RIGHT;
                case '[': offset++; return Tokens.BRACKET_LEFT;
                case ']': offset++; return Tokens.BRACKET_RIGHT;
                case '+': offset++; return Tokens.PLUS;
                case '-': offset++; return Tokens.MINUS;
                case '*': offset++; return Tokens.MULTIPLY;
                case '/': offset++; return Tokens.DIVIDE;
                case '%': offset++; return Tokens.MODULO;
                case '~': offset++; return Tokens.BITWISE_NOT;
                case '|': offset++; return Tokens.BITWISE_OR;
                case '&': offset++; return Tokens.BITWISE_AND;
                case '^': offset++; return Tokens.BITWISE_XOR;
                case '!': offset++; return Tokens.LOGICAL_NOT;
                case '>': offset++; return Tokens.CMP_GT;
                case '<': offset++; return Tokens.CMP_LT;
            }

            throw new Exception("Invalid symbol: " + code.Substring(offset, 3));
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private int MakeReal(decimal value)
        {
            int int_part = decimal.ToInt32(value);
            value = value - int_part;
            int frac_part = decimal.ToInt32(value * 65536);
            return (int_part << 16) | frac_part;
        }
    }
}
