﻿//-----------------------------------------------------------------------------------------
// File:  OpcodeTable.cs
// Title: Opcode Table
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PCC
{
    public enum OpCodes
    {
        NULL, NOP, LABEL, RESET, ENDEX, CALL, RET, JMP, JMP_F, JMP_T, JMP_DZ,
        ADD, SUB, SUBX, IMUL, IDIV, IMOD, RMUL, RDIV, RMOD, INCR, DECR,
        MOV, MOVI, MOVD, MOVS, INIT_ARRAY, INCR_ARRAY,
        BW_AND, BW_OR, BW_XOR, BW_NOT, BW_AND_NOT, BW_TEST, LG_NOT, NEG,
        CMP_EQ, CMP_NE, CMP_GT, CMP_LT, CMP_GE, CMP_LE, CMP_LS, CMP_LC,
        SHIFT_RIGHT, SHIFT_LEFT, ROTATE_RIGHT, ROTATE_LEFT,
        WR_BYTE, WR_WORD, WR_LONG, RD_BYTE, RD_WORD, RD_LONG,
        WAITCNT, WAITPEQ, WAITPNE, WAITVID, LOCKNEW, LOCKSET, LOCKCLR, LOCKRET,
        LIMIT_MIN, LIMIT_MAX, REV, CMPSUB, ABS, ABSN, 
        ITOR, RTOI, SIN, COS, TAN, ASIN, ACOS, LOG, EXP, SQRT, ISQRT,
        CEIL, FLOOR, TRUNC, ROUND, FRACT
    }

    public class OpCodeEntry
    {
        public OpCodes OpCode;
        public VarType Type;
        public int LineNo;
        public ValType Dst;
        public ValType Src;
        public string Name;
        public enum ComparisonT { NONE, EQ, NE, LT, LE, GT, GE, LC, LS };
        public ComparisonT Comparison = ComparisonT.NONE;

        public OpCodeEntry(OpCodes opcode, VarType type, int lineno)
        {
            OpCode = opcode;
            Type = type;
            LineNo = lineno;
        }

        public override string ToString()
        {
            return string.Format("{0,3} [{1} {2}] [{3}] [{4}] {5}", LineNo, OpCode, Type, Dst, Src, Name);
        }
    }

    class OpcodeTable
    {
        private LexAn la;
        
        public List<OpCodeEntry> OpTable;

        public OpcodeTable(LexAn _la)
        {
            la = _la;
            OpTable = new List<OpCodeEntry>();
        }

        public ValType Add(OpCodes opcode, VarType type, ValType dst, ValType src, 
                           OpCodeEntry.ComparisonT comparison = OpCodeEntry.ComparisonT.NONE)
        {
            OpCodeEntry opc = new OpCodeEntry(opcode, type, la.LineNo);
            opc.Dst = dst;
            opc.Src = src;
            opc.Comparison = comparison;
            OpTable.Add(opc);
            return new ValType(OpTable.Count - 1, VarType.OPCODE);
        }

        public ValType Add(OpCodes opcode, VarType type, ValType dst, string name)
        {
            OpCodeEntry opc = new OpCodeEntry(opcode, type, la.LineNo);
            opc.Dst = dst;
            opc.Name = name;
            OpTable.Add(opc);
            return new ValType(OpTable.Count - 1, VarType.OPCODE);
        }

        public ValType Add(OpCodes opcode, VarType type)
        {
            OpCodeEntry opc = new OpCodeEntry(opcode, type, la.LineNo);
            OpTable.Add(opc);
            return new ValType(OpTable.Count - 1, VarType.OPCODE);
        }

        public ValType Add(OpCodes opcode, VarType type, string name)
        {
            OpCodeEntry opc = new OpCodeEntry(opcode, type, la.LineNo);
            opc.Name = name;
            OpTable.Add(opc);
            return new ValType(OpTable.Count - 1, VarType.NONE);
        }

        public void Clear()
        {
            OpTable.Clear();
        }

        public void ClearLineNo()
        {
            OpTable[OpTable.Count - 1].LineNo = 0;
        }

        public int Count { get { return OpTable.Count; } }
        public int PC { get { return OpTable.Count; } }
        public OpCodeEntry this[int index] { get { return OpTable[index]; } }
    }
}
