﻿//-----------------------------------------------------------------------------------------
// File:  Optimizer.cs
// Title: Optimizer
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PCC
{
    class Optimizer
    {
        private OpcodeTable ot;
        private SymbolTable st;

        public Optimizer(OpcodeTable _ot, SymbolTable _st)
        {
            ot = _ot;
            st = _st;
        }

        public void Run()
        {
            RemoveMathResultMoves();
            while (RemoveBackwardMoves() > 0) ;
            RemoveMoveToSelf();
            RemoveRedundantJumps();
            RemoveDoubleJumps();
            RemoveCompareToZero();
            RemoveUnusedLabels();
            FixSubx();
            FixNulls();
            RemoveReturnJumps();
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void FixNulls()
        {
            for (int i = 1; i < ot.Count - 1; i++)
                if (ot[i].OpCode == OpCodes.NULL || ot[i].OpCode == OpCodes.ENDEX)
                    ot[i].LineNo = 0;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void FixSubx()
        {
            int i, j;

            for (i = 1; i < ot.Count - 1; i++)
            {
                j = i + 1;
                if (ot[i].OpCode == OpCodes.SUBX && ot[j].OpCode == OpCodes.MOV)
                {
                    if (ot[j].Src.Type == VarType.OPCODE && ot[j].Src.Value == i)
                    {
                        ot[i].Src = ot[j].Dst;
                        ot[i].Dst = ot[j].Dst;
                        ot[j].OpCode = OpCodes.NULL;
                    }
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveCompareToZero()
        {
            int i, j;

            for (i = 1; i < ot.Count - 1; i++)
            {
                OpCodes cmp = ot[i].OpCode;
                if (cmp == OpCodes.CMP_EQ || cmp == OpCodes.CMP_NE)
                {
                    if (ot[i].Dst.Type == VarType.OPCODE &&
                        st[ot[i].Src.Value].State == SymState.CONST &&
                        st[ot[i].Src.Value].Value == 0)
                    {
                        j = ot[i].Dst.Value;
                        ot[j].Comparison = ot[i].Comparison;
                        ot[i].OpCode = OpCodes.NULL;

                        // Check for TEST instruction
                        if (ot[j].OpCode == OpCodes.BW_AND &&
                            ot[j].Dst.Type == VarType.OPCODE &&
                            ot[ot[j].Dst.Value].OpCode == OpCodes.MOV &&
                            ot[ot[j].Dst.Value].Dst.Type != VarType.OPCODE)
                        {
                            ot[ot[j].Dst.Value].OpCode = OpCodes.NULL;
                            ot[j].OpCode = OpCodes.BW_TEST;
                            ot[j].Dst = ot[ot[j].Dst.Value].Src;
                        }
                    }
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveDoubleJumps()
        {
            int i;

            for (i = 1; i < ot.Count - 1; i++)
            {
                if (ot[i].OpCode == OpCodes.JMP)
                {
                    int prev = i - 1;
                    int next = i + 1;

                    while (ot[prev].OpCode == OpCodes.NULL ||
                           ot[prev].OpCode == OpCodes.LABEL) prev--;
                    while (ot[next].OpCode == OpCodes.NULL) next++;

                    if (ot[prev].OpCode == OpCodes.JMP_T && 
                        ot[next].OpCode == OpCodes.LABEL &&
                        ot[prev].Name == ot[next].Name)
                    {
                        ot[prev].OpCode = OpCodes.JMP_F;
                        ot[prev].Name = ot[i].Name;
                        ot[i].OpCode = OpCodes.NULL;
                    }

                    if (ot[prev].OpCode == OpCodes.JMP_F &&
                        ot[next].OpCode == OpCodes.LABEL &&
                        ot[prev].Name == ot[next].Name)
                    {
                        ot[prev].OpCode = OpCodes.JMP_T;
                        ot[prev].Name = ot[i].Name;
                        ot[i].OpCode = OpCodes.NULL;
                    }
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveReturnJumps()
        {
            int i, j;

            for (i = 1; i < ot.Count - 1; i++)
            {
                if (ot[i].OpCode == OpCodes.JMP)
                {
                    for (j = i + 1; ot[j].OpCode == OpCodes.NULL; j++) ;
                    if (ot[j].OpCode == OpCodes.RET && ot[i].Name == ot[j].Name)
                        ot[i].OpCode = OpCodes.NULL;
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveRedundantJumps()
        {
            int i, p, n;
            OpCodes op;

            for (i = 1; i < ot.Count - 1; i++)
            {
                if (ot[i].OpCode == OpCodes.JMP)
                {
                    for (n = i + 1; ot[n].OpCode == OpCodes.NULL; n++) ;
                    if (ot[n].OpCode != OpCodes.LABEL) continue;

                    for (p = i - 1; ot[p].OpCode == OpCodes.NULL; p--) ;
                    switch (ot[p].OpCode)
                    {
                        case OpCodes.JMP_T:
                            op = OpCodes.JMP_F;
                            break;
                        case OpCodes.JMP_F:
                            op = OpCodes.JMP_T;
                            break;
                        default:
                            continue;
                    }

                    if (ot[p].Name == ot[n].Name)
                    {
                        ot[p].OpCode = op;
                        ot[p].Name = ot[i].Name;
                        ot[i].OpCode = OpCodes.NULL;
                    }
                }
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveMathResultMoves()
        {
            int count = ot.Count;
            int i, j;
            string scope = "";
            int math_r1 = st.Find("math_r1", "").Value;
            int math_r2 = st.Find("math_r2", "").Value;

            for (i = 0; i < count; ++i)
            {
                OpCodeEntry opc = ot[i];
                if (opc.OpCode == OpCodes.LABEL && opc.Name.StartsWith(":") == false)
                    scope = opc.Name;

                if (opc.OpCode == OpCodes.MOV)
                {
                    if (opc.Src.Value == math_r1 || opc.Src.Value == math_r2)
                    {
                        if (IsTempVar(opc.Dst, scope))
                        {
                            j = FindDest(i);
                            if (j >= 0)
                            {
                                ot[j].Src = ot[i].Src;
                                ot[i].OpCode = OpCodes.NULL;
                            }
                        }
                    }
                }
            }
        }

        private int FindDest(int index)
        {
            int i = index;
            while (true)
            {
                i++;
                if (i >= ot.Count) return -1;
                if (ot[i].Src.Type == VarType.OPCODE && ot[i].Src.Value == index) return i;
                if (ot[i].OpCode == OpCodes.ENDEX) return -1;
                if (IsBif(ot[i].OpCode)) return -1;
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveMoveToSelf()
        {
            for (int i = 0; i < ot.Count; i++)
                if (ot[i].OpCode == OpCodes.MOV)
                {
                    ValType src = ot[i].Src;
                    ValType dst = ot[i].Dst;

                    while (src.Type == VarType.OPCODE)
                        src = ot[src.Value].Dst;
                
                    while (dst.Type == VarType.OPCODE)
                        dst = ot[dst.Value].Dst;

                    if (src.Value == dst.Value)
                        ot[i].OpCode = OpCodes.NULL;
                }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void RemoveUnusedLabels()
        {
            List<string> used = new List<string>();
            used.Add(ot[1].Name);

            for (int i = 0; i < ot.Count; i++)
            {
                OpCodes op = ot[i].OpCode;
                if (op == OpCodes.JMP || op == OpCodes.JMP_F || op == OpCodes.JMP_T || op == OpCodes.CALL || op == OpCodes.JMP_DZ)
                    used.Add(ot[i].Name);
            }

            for (int i = 0; i < ot.Count; i++)
                if (ot[i].OpCode == OpCodes.LABEL && ot[i].Name.StartsWith(":"))
                    if (used.Contains(ot[i].Name) == false)
                        ot[i].OpCode = OpCodes.NULL;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private int RemoveBackwardMoves()
        {
            int count = ot.Count;
            int i, j;
            string scope = "";
            int counter = 0;

            for (i = 0; i < count; ++i)
            {
                OpCodeEntry opc = ot[i];
                if (opc.OpCode == OpCodes.LABEL && opc.Name.StartsWith(":") == false)
                    scope = opc.Name;

                j = FindRoot(i, scope);
                if (j >= 0)
                {
                    ot[j].Dst = ot[i].Dst;
                    counter++;
                }
            }
            return counter;
        }

        private int FindRoot(int i, string scope)
        {
            if (IsModifiable(ot[i].OpCode) == false) return -1;

            ValType src = ot[i].Src;
            ValType dst = ot[i].Dst;

            if (dst.Type != VarType.ANY &&
                dst.Type != VarType.AINT &&
                dst.Type != VarType.UINT && 
                dst.Type != VarType.INT && 
                dst.Type != VarType.REAL) return -1;
            if (IsTempVar(dst, scope) == true) return -1;
            if (src.Type != VarType.OPCODE) return -1;
            if (IsHWReg(dst) == true) return -1;

            int j = src.Value;
            while (ot[j].Dst.Type == VarType.OPCODE)
                j = ot[j].Dst.Value;

            if (IsModifiable(ot[j].OpCode) == false) return -1;
            if (IsTempVar(ot[j].Dst, scope) == false) return -1;

            while (true)
            {
                if (ot[i].Src.Equals(dst)) return -1;
                if (--i <= j) break;
                if (ot[i].Dst.Equals(dst)) return -1;
            }

            return j;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private bool IsAssign(OpCodes op)
        {
            switch (op)
            {
                case OpCodes.MOV: return true;
                case OpCodes.ADD: return true;
                case OpCodes.SUB: return true;
                case OpCodes.IMUL: return true;
                case OpCodes.IDIV: return true;
                case OpCodes.IMOD: return true;
                case OpCodes.BW_AND: return true;
                case OpCodes.BW_OR: return true;
                case OpCodes.BW_AND_NOT: return true;
                case OpCodes.BW_XOR: return true;
            }
            return false;
        }

        private bool IsModifiable(OpCodes op)
        {
            switch (op)
            {
                case OpCodes.MOV: return true;
                case OpCodes.RD_LONG: return true;
                case OpCodes.RD_WORD: return true;
                case OpCodes.RD_BYTE: return true;
                case OpCodes.ABS: return true;
                case OpCodes.ABSN: return true;
                case OpCodes.FLOOR: return true;
                case OpCodes.CEIL: return true;
                case OpCodes.NEG: return true;
            }
            return false;
        }

        private bool IsHWReg(ValType v)
        {
            string reg = st[v.Value].Name;
            switch (reg)
            {
                case "cnt": return true;
                case "outa": return true;
                case "outb": return true;
                case "dira": return true;
                case "dirb": return true;
                case "par": return true;
                case "ctra": return true;
                case "ctrb": return true;
                case "frqa": return true;
                case "frqb": return true;
                case "vcfg": return true;
                case "vscl": return true;
                case "ina": return true;
                case "inb": return true;
            }
            return false;
        }

        private bool IsBif(OpCodes op)
        {
            switch (op)
            {
                case OpCodes.IMUL:
                case OpCodes.RMUL:
                case OpCodes.IDIV:
                case OpCodes.RDIV:
                case OpCodes.IMOD:
                case OpCodes.RMOD:
                case OpCodes.SIN:
                case OpCodes.COS:
                case OpCodes.TAN:
                case OpCodes.ASIN:
                case OpCodes.ACOS:
                case OpCodes.LOG:
                case OpCodes.EXP:
                case OpCodes.SQRT:
                case OpCodes.ISQRT:
                case OpCodes.CALL:
                    return true;
            }
            return false;
        }

        private bool IsTempVar(ValType v, string scope)
        {
            if (v.Type == VarType.OPCODE || v.Type == VarType.VOID)
                return false;

            scope += "__";
            string s = st[v.Value].Name;
            if (s.Length < scope.Length) return false;
            if (s.Substring(0, scope.Length) != scope) return false;
            return true;
        }

        private bool IsLocalVar(ValType v, string scope)
        {
            if (v.Type == VarType.OPCODE || v.Type == VarType.VOID)
                return false;

            scope += "_";
            string s = st[v.Value].Name;
            if (s.Length < scope.Length) return false;
            if (s.Substring(0, scope.Length) != scope) return false;
            return true;
        }
    }
}
