﻿//-----------------------------------------------------------------------------------------
// File:  Parser.cs
// Title: Parser
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------------------------------------------------------
// <>   Rule name
// ::=  Is defined by
// |    Or
// []   Optional
// {}   Repeat one or more times
// {}*  Repeat zero or more times
// ()   Group
// ''   Literal symbol
// ABC  Keyword
// ;    Rule terminator
//-------------------------------------------------------------------------------------------------------------------------------------------------
// <program> ::= { <data-decl-stmt> | <func-decl> }* ;
//
// <func-decl> ::= <return-data-type> <identifier> '(' [ <func-parms> ] ')' <func-body> ;
//
// <func-parms> ::= <data-type> <identifier> { ',' <data-type> <identifier> }* ;
//
// <return-data-type> ::= <data-type> | VOID ;
//
// <func-body> ::= '{' { <data-decl-stmt> }* { <statement> } '}' ;
//
// <data-decl-stmt> ::= <data-type> <data-decl> { ',' <data-decl> }* ';' ;
//
// <data-decl> := <identifier> [ '=' [ '+' | '-' ] <constant> ] ;
//
// <data-type> ::= INT | UINT | REAL ;
//
// <statement> ::= <assignment-stmt>
//               | <function-stmt>
//               | <dot-assign-stmt>
//               | <if-stmt>
//               | <while-stmt>
//               | <do-while-stmt>
//               | <for-stmt>
//               | <switch-stmt>
//               | <break-stmt>
//               | <continue-stmt>
//               | <return-stmt>
//               | <compound-stmt>
//               | <global-array-stmt>
//               | <wait-count-stmt>
//               | <wait-stmt>
//               | <lock-stmt>
//               | <built-in-op-stmt>
//               | <nop-stmt>
//               | ';' ;
//
// <assignment-stmt> ::= <assignment> ';' ;
//
// <assignment> ::= <identifier> <assigmment-op> <expression>
//                | <identifier> '=' [TRUE | FALSE] ;
//
// <assignment-op> ::= '=' | '+=' | '-=' | '*=' | '/=' | '%=' | '|=' | '&=' | '^=' | '&~=' | '<<=' | '>>=' | '<-=' | '->=' ;
//
// <dot-assign-stmt> ::= <identifier> ( '.I' | '.D' | '.S' ) '=' <expression> ';' ;
//
// <global-array-stmt> ::= <global-array> '=' <expression> ';' ;
//
// <global-array> := ( GBYTE | GWORD | GLONG ) '[' <expression> ']' ;
//
// <if-stmt> ::= IF '(' <logical-or-expression> ')' <statement> [ ELSE <statement> ] ;
//
// <while-stmt> ::= WHILE '(' <logical-or-expression> ')' <statement> ;
//
// <do-while-stmt> ::= DO <statement> WHILE '(' <logical-or-expression> ')' ';' ;
//
// <for-stmt> ::= FOR '(' <for-init> [ ';' <logical-or-expression> ';' <for-incr> ] ')' <statement> ;
//
// <for-init> ::= <identifier> '=' <expression> ;
//
// <for-incr> ::= <identifier> <assignment-op> <expression> ;
//
// <switch-stmt> ::= SWITCH '(' <expression> ')' '{' { <case-clause> }* [<default-clause>] '}' ;
//
// <case-clause> ::= CASE <integer-constant> ':' { <statement> }* BREAK ';' ;
//
// <default-clause> ::= DEFAULT ':' { <statement> }* BREAK ';' ;
//
// <break-stmt> ::= BREAK ';' ;
//
// <continue-stmt> ::= CONTINUE ';' ;
//
// <return-stmt> ::= RETURN [<expression>] ';' ;
//
// <function-stmt> ::= <function-call> ';' ;
//
// <function-call> ::= <identifier> '(' [ <expression> { ',' <expression>} }* ] ')' ;
//
// <wait-count-stmt> ::= WAITCNT '(' <identifier> ',' <expression> ')' ';'
//                     | WAITCNT '(' <expression> ')' ';' ;
//
// <wait-stmt> ::= ( WAITPEQ | WAITPNE | WAITVID ) '(' <expression> ',' <expression> ')' ';' ;
//
// <lock-stmt> ::= ( LOCKNEW | LOCKSET | LOCKCLR | LOCKRET ) '(' <IDENTIFIER> ')' ';' ;
//
// <built-in-op-stmt> ::= ( CMPSUB | REV ) '(' <identifier> ',' <expression> ')' ';' ;
//
// <compound-stmt> ::= '{' { <statement> } '}' ;
//
// <nop-stmt> ::= NOP ';' ;
//
// <logical-or-expression> ::= <logical-and-expression>
//                           | <logical-or-expression '||' <logical-and-expression> ;
//
// <logical-and-expression> ::= <relational-expression>
//                            | <logical-and-expression '&&' <relational-expression> ;
//
// <relational-expression> ::= <expression> '==' <expression>
//                           | <expression> '!=' <expression>
//                           | <expression> '<' <expression>
//                           | <expression> '>' <expression>
//                           | <expression> '<=' <expression>
//                           | <expression> '>=' <expression>
//                           | <lock-expression>
//                           | <parity-expression> ;
//
// <lock-expression> ::= ( LOCKSET | LOCKCLR ) '(' <identifier> ')' ( '==' | '!=' } ( TRUE | FALSE ) ;
//
// <parity-expression> ::= <parity-function> ( '==' | '!=' } ( ODD | EVEN ) ;
//
// <expression> ::= <inclusive-or-expression>
//                | <expression> '#>' <inclusive-or-expression>
//                | <expression> '<#' <inclusive-or-expression> ;
//
// <inclusive-or-expression> ::= <exclusive-or-expression>
//                             | <inclusive-or-expression> '|' <exclusive-or-expression> ;
//
// <exclusive-or-expression> ::= <and-expression>
//                             | <exclusive-or-expression> '^' <and-expression> ;
//
// <and-expression> ::= <shift-expression>
//                    | <and-expression> '&' <shift-expression>
//                    | <and-expression> '&~' <shift-expression> ;
//
// <shift-expression> ::= <additive-expression>
//                      | <shift-expression> '<<' <additive-expression>
//                      | <shift-expression> '>>' <additive-expression>
//                      | <shift-expression> '->' <additive-expression>
//                      | <shift-expression> '<-' <additive-expression> ;
//
// <additive-expression> ::= <multiplicative-expression>
//                         | <additive-expression> '+' <multiplicative-expression>
//                         | <additive-expression> '-' <multiplicative-expression> ;
//
// <multiplicative-expression> ::= <unary-expression>
//                               | <multiplicative-expression> '*' <unary-expression>
//                               | <multiplicative-expression> '/' <unary-expression>
//                               | <multiplicative-expression> '%' <unary-expression> ;
//
// <unary-expression> ::= <primary-expression>
//                      | ! <primary-expression>
//                      | ~ <primary-expression>
//                      | - <primary-expression>
//                      | + <primary-expression> ;
//
// <primary-expression> ::= <identifier>
//                        | <function-call>
//                        | <global-array>
//                        | <built-in-function>
//                        | <parity-function>
//                        | <built-in-operation>
//                        | <integer-constant>
//                        | <real-constant>
//                        | '(' <expression> ')' ;
//
// <built-in-function> ::= ( SIN | COS | TAN | ASIN | ACOS | LOG | EXP | SQRT | ISQRT ) '(' <expression> ')' ;
//
// <built-in-operation> ::= ( ABS | ABSN | ITOR | RTOI | CEIL | FLOOR | TRUNC | ROUND | FRACT ) '(' <expression> ')' ;
//
// <parity-function> ::= PARITY '(' <expression> ',' <expression> ')' ;
//
// <identifier> ::= <letter> { <letter> | <digit> | '_' }* ;
//
// <letter> ::= (a-zA-Z) ;
//
// <digit> ::= (0-9) ;
//
// <constant> ::= <integer-constant> | <real-constant> ;
//
// <real-constant> ::= <dec-constant> '.' [ <dec-constant> ] ;
//
// <integer-constant> ::= <dec-constant> | <hex-constant> | <oct-constant> | <bin-constant> ;
//
// <dec-constant> ::= { 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | _ } ;
//
// <hex-constant> ::= '0x' { 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | A | B | C | D | E | F | _ } ;
//
// <oct-constant> ::= '0o' { 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | _ } ;
//
// <bin-constant> ::= '0b' { 0 | 1 | _ } ;
//-------------------------------------------------------------------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

//-------------------------------------------------------------------------------------------------------------------------------------------------
namespace PCC
{
    class Parser
    {
        private struct FuncState
        {
            public string Name;
            public int State;
            public FuncState(string name, int state)
            {
                Name = name;
                State = state;
            }
        }
        private List<FuncState> FuncList = new List<FuncState>();

        private struct LabelSet
        {
            public string Start;
            public string Finish;
            public LabelSet(string start, string finish)
            {
                Start = start;
                Finish = finish;
            }
        }
        private Stack<LabelSet> LabelSets = new Stack<LabelSet>();

        private LexAn la = null;
        private FuncTable ft = null;
        private OpcodeTable ot = null;
        private SymbolTable st = null;
        private string errmsg = "";
        private string label = "";
        private string func_name = "";
        private VarType func_type;
        private int labelid;

        //-----------------------------------------------------------------------------------------------------------------------------------------
        //-----------------------------------------------------------------------------------------------------------------------------------------
        public Parser(FuncTable _ft, OpcodeTable _ot, SymbolTable _st, LexAn _la)
        {
            ft = _ft;
            ot = _ot;
            st = _st;
            la = _la;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        //-----------------------------------------------------------------------------------------------------------------------------------------
        public void Parse(string filename)
        {
            try
            {
                labelid = 1;
                errmsg = "";

                st.Clear();
                st.Add("dira", VarType.UINT);
                st.Add("dirb", VarType.UINT);
                st.Add("ina", VarType.UINT);
                st.Add("inb", VarType.UINT);
                st.Add("outa", VarType.UINT);
                st.Add("outb", VarType.UINT);
                st.Add("par", VarType.UINT);
                st.Add("cnt", VarType.UINT);
                st.Add("ctra", VarType.UINT);
                st.Add("ctrb", VarType.UINT);
                st.Add("frqa", VarType.UINT);
                st.Add("frqb", VarType.UINT);
                st.Add("phsa", VarType.UINT);
                st.Add("phsb", VarType.UINT);
                st.Add("vcfg", VarType.UINT);
                st.Add("vscl", VarType.UINT);
                st.Add("tru", VarType.AINT, 1);
                st.Add("fls", VarType.AINT, 0);
                st.Add("even", VarType.AINT, 0);
                st.Add("odd", VarType.AINT, 1);

                st.Add("math_p1", "", VarType.ANY);
                st.Add("math_p2", "", VarType.ANY);
                st.Add("math_r1", "", VarType.ANY);
                st.Add("math_r2", "", VarType.ANY);
                st.Add("math_t", "", VarType.ANY);
                st.Add("math_s", "", VarType.ANY);
                st.Add("math_ones", "", VarType.ANY, -1);
                st.Add("math_real_mask", "", VarType.ANY, 0xFFFF);
                st.Add("math_half", "", VarType.ANY, 0x8000);
                st.Add("math_90", "", VarType.REAL, 0x00800000);
                st.Add("math_180", "", VarType.REAL, 0x01000000);
                st.Add("0", VarType.AINT, 0);

                ot.Clear();
                ft.Clear();
                FuncList.Clear();
                
                program();
            }
            catch (Exception ex)
            {
                errmsg = string.Format("ERROR: {0}: At line {1}\n", ex.Message, la.LineNo);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        //-----------------------------------------------------------------------------------------------------------------------------------------
        public string ErrorMsg { get { return errmsg; } }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <program> ::= { <data-decl-stmt> | <func-decl> }* ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void program()
        {
            string label;

            do
            {
                func_type = la.Var_Type;
                la.Next(Tokens.VAR_TYPE); // TYPE

                label = la.Label;
                la.Next(Tokens.LABEL); // LABEL

                if (la.Peek == Tokens.PAREN_LEFT)
                {
                    func_name = label;
                    func_decl();
                }
                else
                {
                    func_name = "";
                    data_decl_stmt(func_type, label);
                }
            }
            while (la.Peek != Tokens.NONE);

            // Second pass, process function bodies
            foreach (FuncState fs in FuncList)
            {
                func_name = fs.Name;
                la.State = fs.State;
                func_body();
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <func-decl> ::= <return-data-type> <identifier> '(' [ <data-type> <identifier> { ',' <data-type> <identifier> }* ] ')' <func-body> ;
        // <return-data-type> ::= <data-type> | VOID ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void func_decl()
        {
            if (func_type != VarType.VOID && func_type != VarType.INT && func_type != VarType.UINT && func_type != VarType.REAL)
                throw new Exception("Not a valid function type");

            ft.AddFunc(func_name, func_type);

            if (func_type != VarType.VOID)
                st.Add(func_name + '_', "", func_type);

            la.Next(Tokens.PAREN_LEFT);

            if (la.Peek != Tokens.PAREN_RIGHT)
            {
                while (true)
                {
                    if (la.Peek != Tokens.VAR_TYPE)
                        throw new Exception("Expected a data type");
                    VarType type = la.Var_Type;
                    if (type != VarType.INT && type != VarType.UINT && type != VarType.REAL)
                        throw new Exception("Invalid data type for a function parameter");
                    la.Next(Tokens.VAR_TYPE);

                    if (la.Peek != Tokens.LABEL)
                        throw new Exception("Expected a data label");
                    int index = st.Add(la.Label, func_name, type);
                    ft.AddParm(new ValType(index, type));
                    la.Next(Tokens.LABEL);

                    if (la.Peek == Tokens.PAREN_RIGHT) break;
                    la.Next(Tokens.COMMA);
                }
            }

            la.Next(Tokens.PAREN_RIGHT);
            FuncList.Add(new FuncState(func_name, la.State));

            if (la.Peek != Tokens.BRACE_LEFT)
                throw new Exception("Expected a left parenthesis");

            // Skip function body
            int brace_count = 1;
            while (brace_count > 0 && la.Peek != Tokens.NONE)
            {
                la.Next(); // func body
                if (la.Peek == Tokens.BRACE_RIGHT) brace_count--;
                if (la.Peek == Tokens.BRACE_LEFT) brace_count++;
            }
            la.Next(Tokens.BRACE_RIGHT);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <func-body> ::= '{' { <data-decl-stmt> }* { <statement> } '}' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void func_body()
        {
            string label;
            VarType type;

            ot.Add(OpCodes.LABEL, VarType.NONE, func_name);

            la.Next(Tokens.BRACE_LEFT);

            while (la.Peek == Tokens.VAR_TYPE)
            {
                type = la.Var_Type;
                la.Next(Tokens.VAR_TYPE);

                label = la.Label;
                la.Next(Tokens.LABEL);

                data_decl_stmt(type, label);
            }

            while (la.Peek != Tokens.BRACE_RIGHT)
            {
                statement();
            }

            ot.Add(OpCodes.RET, VarType.NONE, func_name + "_RET");
            la.Next(Tokens.BRACE_RIGHT);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <data-decl-stmt> ::= <data-type> <data-decl> { ',' <data-decl> }* ';' ;
        // <data-decl> := <identifier> [ '=' [ '+' | '-' ] <constant> ] ;
        // <data-type> ::= INT | UINT | REAL ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void data_decl_stmt(VarType type, string label)
        {
            int const_value;
            string const_label;

            if (type != VarType.INT && type != VarType.UINT && type != VarType.REAL)
                throw new Exception("Invalid data type type");

            while (true)
            {
                if (la.Peek == Tokens.ASSIGN && la.AssignOp == Tokens.ASSIGN)
                {
                    la.Next(Tokens.ASSIGN);

                    if (la.Peek == Tokens.PLUS)
                        la.Next(Tokens.PLUS);

                    if (la.Peek == Tokens.MINUS)
                    {
                        la.Next(Tokens.MINUS);
                        const_value = -la.Value;
                        const_label = "M_" + la.Label;
                    }
                    else
                    {
                        const_value = la.Value;
                        const_label = la.Label;
                    }


                    if (type == VarType.INT || type == VarType.UINT)
                        la.Next(Tokens.CONSTANT);
                    else if (type == VarType.REAL)
                        la.Next(Tokens.REAL_CONSTANT);
                    else
                        throw new Exception("Invalid constant type");

                    if (func_name == "")
                        st.Add(label, func_name, type, const_value);
                    else
                    {
                        ValType dst = new ValType(st.Add(label, func_name, type), type);
                        ValType src = AddConstant(type, const_value, const_label);
                        ot.Add(OpCodes.MOV, type, dst, src);
                    }
                }
                else
                    st.Add(label, func_name, type);

                if (la.Peek == Tokens.SEMICOLON) break;
                la.Next(Tokens.COMMA);

                label = la.Label;
                la.Next(Tokens.LABEL);
            }

            la.Next(Tokens.SEMICOLON);
        }
        
        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <statement> ::= <assignment-stmt>
        //               | <function-stmt>
        //               | <dot-assign-stmt>
        //               | <if-stmt>
        //               | <while-stmt>
        //               | <do-while-stmt>
        //               | <for-stmt>
        //               | <switch-stmt>
        //               | <break-stmt>
        //               | <continue-stmt>
        //               | <return-stmt>
        //               | <compound-stmt>
        //               | <global-array-stmt>
        //               | <wait-count-stmt>
        //               | <wait-stmt>
        //               | <lock-stmt>
        //               | <built-in-op-stmt>
        //               | <nop-stmt>
        //               | ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void statement()
        {
            switch (la.Peek)
            {
                case Tokens.LABEL:
                    label = la.Label;
                    la.Next(Tokens.LABEL);
                    switch (la.Peek)
                    {
                        case Tokens.ASSIGN:
                            assignment_stmt(label);
                            break;
                        case Tokens.PAREN_LEFT:
                            function_stmt(label);
                            break;
                        case Tokens.DOTI:
                        case Tokens.DOTS:
                        case Tokens.DOTD:
                            dot_assign_stmt(label);
                            break;
                        default:
                            throw new Exception("Expected assignment or function call");
                    }
                    break;
                case Tokens.IF:
                    if_stmt();
                    break;
                case Tokens.WHILE:
                    while_stmt();
                    break;
                case Tokens.DO:
                    do_stmt();
                    break;
                case Tokens.FOR:
                    for_stmt();
                    break;
                case Tokens.SWITCH:
                    switch_stmt();
                    break;
                case Tokens.BREAK:
                    break_stmt();
                    break;
                case Tokens.CONTINUE:
                    continue_stmt();
                    break;
                case Tokens.RETURN:
                    return_stmt();
                    break;
                case Tokens.BRACE_LEFT:
                    compound_stmt();
                    break;
                case Tokens.GBYTE:
                case Tokens.GWORD:
                case Tokens.GLONG:
                    global_array_stmt();
                    break;
                case Tokens.WAITCNT:
                    wait_count_stmt();
                    break;
                case Tokens.WAITPEQ:
                case Tokens.WAITPNE:
                case Tokens.WAITVID:
                    wait_stmt();
                    break;
                case Tokens.LOCKNEW:
                case Tokens.LOCKSET:
                case Tokens.LOCKCLR:
                case Tokens.LOCKRET:
                    lock_stmt();
                    break;
                case Tokens.BIOP:
                    biop_stmt();
                    break;
                case Tokens.NOP:
                    nop_stmt();
                    break;
                case Tokens.SEMICOLON:
                    la.Next(Tokens.SEMICOLON);
                    break;
                default:
                    throw new Exception("Expected start of a valid statement");
            }

            st.ReleaseTemps();
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <compound-stmt> ::= '{' { <statement> } '}' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void compound_stmt()
        {
            la.Next(Tokens.BRACE_LEFT);
            while (la.Peek != Tokens.BRACE_RIGHT)
            {
                statement();
            }
            la.Next(Tokens.BRACE_RIGHT);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <function-stmt> ::= <function-call> ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void function_stmt(string label)
        {
            function_call(label);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <function-call> ::= <identifier> '(' [ <expression> { ',' <expression>} }* ] ')' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType function_call(string label)
        {
            if (label == func_name)
                throw new Exception("Reentrant functions are not supported");
 
            FuncEntry fe = ft[label];
            ValType rval = new ValType();

            la.Next(Tokens.PAREN_LEFT);

            if (la.Peek != Tokens.PAREN_RIGHT)
            {
                int index = 0;
                while (true)
                {
                    ValType parm1 = fe.ParmList[index++];
                    ValType parm2 = expression(true);
                    ot.Add(OpCodes.MOV, CheckTypes(parm1, parm2), parm1, parm2);

                    if (la.Peek != Tokens.COMMA) break;
                    la.Next(Tokens.COMMA);
                }
            }

            ot.Add(OpCodes.CALL, fe.Type, label);
            if (fe.Type != VarType.VOID)
                rval = st.Find(label + '_', "");

            la.Next(Tokens.PAREN_RIGHT);

            return rval;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <if-stmt> ::= IF '(' <logical-or-expression> ')' <statement> [ ELSE <statement> ] ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void if_stmt()
        {
            string true_label = GetLabel();
            string false_label = GetLabel();
            string skip_label = GetLabel();
            
            la.Next(Tokens.IF);
            la.Next(Tokens.PAREN_LEFT);

            logical_or_expression(ref true_label, ref false_label);
            ot.Add(OpCodes.JMP_F, VarType.NONE, false_label);

            la.Next(Tokens.PAREN_RIGHT);

            ot.Add(OpCodes.LABEL, VarType.NONE, true_label);
            statement();

            if (la.Peek == Tokens.ELSE)
            {
                la.Next(Tokens.ELSE);

                ot.Add(OpCodes.JMP, VarType.NONE, skip_label);
                ot.Add(OpCodes.LABEL, VarType.NONE, false_label);

                statement();

                ot.Add(OpCodes.LABEL, VarType.NONE, skip_label);
                ot.ClearLineNo();
            }
            else
            {
                ot.Add(OpCodes.LABEL, VarType.NONE, false_label);
                ot.ClearLineNo();
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <while-stmt> ::= WHILE '(' <logical-or-expression> ')' <statement> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void while_stmt()
        {
            bool true_flag = false;
            string start_label = GetLabel();
            string true_label = GetLabel();
            string false_label = GetLabel();

            la.Next(Tokens.WHILE);
            la.Next(Tokens.PAREN_LEFT);

            ot.Add(OpCodes.LABEL, VarType.NONE, start_label);

            if (la.Peek == Tokens.LABEL && la.Label == "CONST_tru")
            {
                la.Next(Tokens.LABEL);
                true_flag = true;
            }
            else
            {
                logical_or_expression(ref true_label, ref false_label);
            }

            la.Next(Tokens.PAREN_RIGHT);

            if (la.Peek == Tokens.SEMICOLON)
            {
                ot.Add(OpCodes.JMP_T, VarType.NONE, start_label);
                ot.Add(OpCodes.LABEL, VarType.NONE, true_label);
                la.Next(Tokens.SEMICOLON);
            }
            else
            {
                if (true_flag == false)
                    ot.Add(OpCodes.JMP_F, VarType.NONE, false_label);
                ot.Add(OpCodes.LABEL, VarType.NONE, true_label);

                LabelSets.Push(new LabelSet(start_label, false_label));
                statement();
                LabelSets.Pop();

                ot.Add(OpCodes.JMP, VarType.NONE, start_label);
                ot.ClearLineNo();
            }
            ot.Add(OpCodes.LABEL, VarType.NONE, false_label);
            ot.ClearLineNo();
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <do-while-stmt> ::= DO <statement> WHILE '(' <logical-or-expression> ')' ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void do_stmt()
        {
            string true_label = GetLabel();
            string false_label = GetLabel();

            ot.Add(OpCodes.LABEL, VarType.NONE, true_label);

            la.Next(Tokens.DO);

            LabelSets.Push(new LabelSet(true_label, false_label));
            statement();
            LabelSets.Pop();

            la.Next(Tokens.WHILE);
            la.Next(Tokens.PAREN_LEFT);

            logical_or_expression(ref true_label, ref false_label);

            ot.Add(OpCodes.JMP_T, VarType.NONE, true_label);
            ot.Add(OpCodes.LABEL, VarType.NONE, false_label);

            la.Next(Tokens.PAREN_RIGHT);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <break-stmt> ::= BREAK ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void break_stmt()
        {
            if (LabelSets.Count == 0)
                throw new Exception("Break statement must be inside a while, do, for or switch statement");

            string label = LabelSets.Peek().Finish;
            ot.Add(OpCodes.JMP, VarType.NONE, label);

            la.Next(Tokens.BREAK);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <continue-stmt> ::= CONTINUE ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void continue_stmt()
        {
            if (LabelSets.Count == 0)
                throw new Exception("Continue statement must be inside a while, do or for statement");

            string label = LabelSets.Peek().Start;
            ot.Add(OpCodes.JMP, VarType.NONE, label);

            la.Next(Tokens.CONTINUE);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <for-stmt> ::= FOR '(' <for-init> [ ';' <logical-or-expression> ';' <for-incr> ] ')' <statement> ;
        // <for-init> ::= <identifier> '=' <expression> ;
        // <for-incr> ::= <assignment> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void for_stmt()
        {
            string label;
            string start_label = GetLabel();
            string exit_label = GetLabel();
            int state1, state2;

            la.Next(Tokens.FOR);
            la.Next(Tokens.PAREN_LEFT);

            label = la.Label;
            la.Next(Tokens.LABEL);
            ValType dst = st.Find(label, func_name);

            la.Next(Tokens.ASSIGN);

            ValType src = expression(true);
            ot.Add(OpCodes.MOV, CheckTypes(dst, src), dst, src);

            if (la.Peek == Tokens.PAREN_RIGHT)
            {
                int ln = la.LineNo;
                la.Next(Tokens.PAREN_RIGHT);

                ot.Add(OpCodes.LABEL, VarType.NONE, start_label);

                LabelSets.Push(new LabelSet(start_label, exit_label));
                statement();
                LabelSets.Pop();

                ot.Add(OpCodes.JMP_DZ, VarType.NONE, dst, start_label);
                ot.ClearLineNo();
                ot.Add(OpCodes.LABEL, VarType.NONE, exit_label);
                ot.ClearLineNo();

                return;
            }

            ot.Add(OpCodes.LABEL, VarType.NONE, start_label);

            la.Next(Tokens.SEMICOLON);

            string true_label = GetLabel();
            string false_label = GetLabel();
            logical_or_expression(ref true_label, ref false_label);
            ot.Add(OpCodes.JMP_F, VarType.NONE, false_label);
            ot.Add(OpCodes.LABEL, VarType.NONE, true_label);

            la.Next(Tokens.SEMICOLON);
            state1 = la.State;

            // Skip increment statement
            while (la.Peek != Tokens.PAREN_RIGHT) la.Next();
            state2 = la.State;

            la.Next(Tokens.PAREN_RIGHT);

            LabelSets.Push(new LabelSet(start_label, false_label));
            statement();
            LabelSets.Pop();

            // Insert increment statement
            la.State = state1;
            label = la.Label;
            la.Next(Tokens.LABEL);
            assignment(label);

            la.State = state2;
            int brace = 0;
            while (true)
            {
                la.Next();
                if (la.Peek == Tokens.NONE) break;
                if (la.Peek == Tokens.SEMICOLON && brace == 0) break;
                if (la.Peek == Tokens.BRACE_LEFT) brace++;
                if (la.Peek == Tokens.BRACE_RIGHT)
                {
                    brace--;
                    if (brace == 0) break;
                }
            }

            ot.Add(OpCodes.JMP, VarType.NONE, start_label);
            ot.Add(OpCodes.LABEL, VarType.NONE, false_label);

            la.Next(); // ';' or '}'
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <switch-stmt> ::= SWITCH '(' <expression> ')' '{' { <case-clause> }* [<default-clause>] '}' ;
        // <case-clause> ::= CASE <integer-constant> ':' { <statement> }* BREAK ';' ;
        // <default-clause> ::= DEFAULT ':' { <statement> }* BREAK ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void switch_stmt()
        {
            string exit_label = GetLabel();

            la.Next(Tokens.SWITCH);
            la.Next(Tokens.PAREN_LEFT);

            ValType dst = expression(true);

            la.Next(Tokens.PAREN_RIGHT);
            la.Next(Tokens.BRACE_LEFT);

            while (la.Peek == Tokens.CASE)
            {
                int offset;
                string jmp_label = GetLabel();

                la.Next(Tokens.CASE);

                if (la.Peek == Tokens.CONSTANT)
                {
                    offset = st.Contains(la.Label, func_name);
                    if (offset < 0)
                        offset = st.Add(la.Label, VarType.AINT, la.Value);
                    la.Next(Tokens.CONSTANT);
                }
                else
                {
                    offset = st.Contains(la.Label, "");
                    if (offset < 0 || st[offset].State != SymState.CONST)
                        throw new Exception("Expected an integer constant");
                    la.Next(Tokens.LABEL);
                }

                ValType src = new ValType(offset, dst.Type);

                ot.Add(OpCodes.CMP_NE, dst.Type, dst, src, OpCodeEntry.ComparisonT.NE);
                ot.Add(OpCodes.JMP_T, VarType.NONE, jmp_label);

                la.Next(Tokens.COLON);

                while (la.Peek != Tokens.BREAK && la.Peek != Tokens.NONE)
                    statement();

                ot.Add(OpCodes.JMP, VarType.NONE, exit_label);
                ot.Add(OpCodes.LABEL, VarType.NONE, jmp_label);

                la.Next(Tokens.BREAK);
                la.Next(Tokens.SEMICOLON);
            }

            if (la.Peek == Tokens.DEFAULT)
            {
                la.Next(Tokens.DEFAULT);
                la.Next(Tokens.COLON);

                while (la.Peek != Tokens.BREAK && la.Peek != Tokens.NONE)
                    statement();

                la.Next(Tokens.BREAK);
                la.Next(Tokens.SEMICOLON);
            }

            ot.Add(OpCodes.LABEL, VarType.NONE, exit_label);

            la.Next(Tokens.BRACE_RIGHT);
        }
        
        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <return-stmt> ::= RETURN [<expression>] ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void return_stmt()
        {
            la.Next(Tokens.RETURN);

            if (la.Peek != Tokens.SEMICOLON)
            {
                ValType rval = expression(true);
                ValType lval = st.Find(func_name + '_', "");
                ot.Add(OpCodes.MOV, CheckTypes(lval, rval), lval, rval);
            }

            ot.Add(OpCodes.JMP, VarType.NONE, func_name + "_RET");

            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <assignment-stmt> ::= <assigmment> ';'
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void assignment_stmt(string label)
        {
            assignment(label);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <assignment> ::= <identifier> <assigmment-op> <expression>
        //                | <identifier> '=' [TRUE | FALSE] ;
        // <assignment-op> ::= '=' | '+=' | '-=' | '*=' | '/=' | '%=' | '|=' | '&=' | '^=' | '&~=' | '<<=' | '>>=' | '<-=' | '->=' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void assignment(string label)
        {
            string result = "";
            ValType op1 = st.Find(label, func_name);
            VarType rtype = op1.Type;
            Tokens op = la.AssignOp;
            OpCodes opc = OpCodes.NOP;

            la.Next(Tokens.ASSIGN);

            ValType op2 = expression(false);

            switch (op)
            {
                case Tokens.ASSIGN: opc = OpCodes.MOV; break;
                case Tokens.PLUS: opc = OpCodes.ADD; break;
                case Tokens.MINUS: opc = OpCodes.SUB; break;
                case Tokens.MULTIPLY: opc = OpCodes.IMUL; result = "math_r1"; break;
                case Tokens.DIVIDE: opc = OpCodes.IDIV; result = "math_r1"; break;
                case Tokens.MODULO: opc = OpCodes.IMOD; result = "math_r2"; break;
                case Tokens.BITWISE_OR: opc = OpCodes.BW_OR; break;
                case Tokens.BITWISE_AND: opc = OpCodes.BW_AND; break;
                case Tokens.BITWISE_XOR: opc = OpCodes.BW_XOR; break;
                case Tokens.BITWISE_AND_NOT: opc = OpCodes.BW_AND_NOT; break;
                case Tokens.SHIFT_LEFT: opc = OpCodes.SHIFT_LEFT; break;
                case Tokens.SHIFT_RIGHT: opc = OpCodes.SHIFT_RIGHT; break;
                case Tokens.ROTATE_LEFT: opc = OpCodes.ROTATE_LEFT; break;
                case Tokens.ROTATE_RIGHT: opc = OpCodes.ROTATE_RIGHT; break;
            }

            switch (op)
            {
                case Tokens.ASSIGN:
                case Tokens.PLUS:
                case Tokens.MINUS:
                case Tokens.MULTIPLY:
                case Tokens.DIVIDE:
                case Tokens.MODULO:
                    CheckTypes(op2, new ValType(0, rtype));
                    break;
                case Tokens.BITWISE_OR:
                case Tokens.BITWISE_AND:
                case Tokens.BITWISE_XOR:
                case Tokens.BITWISE_AND_NOT:
                case Tokens.SHIFT_LEFT:
                case Tokens.SHIFT_RIGHT:
                case Tokens.ROTATE_LEFT:
                case Tokens.ROTATE_RIGHT:
                    CheckTypes(op2, new ValType(0, VarType.AINT));
                    break;
            }

            if (opc == OpCodes.IMUL || opc == OpCodes.IDIV || opc == OpCodes.IMOD)
            {
                ot.Add(OpCodes.MOV, op1.Type, st.Find("math_p1", ""), op1);
                ot.Add(OpCodes.MOV, op2.Type, st.Find("math_p2", ""), op2);
                ot.Add(opc, op1.Type, op1, new ValType());
                ot.Add(OpCodes.MOV, op1.Type, op1, st.Find(result, ""));
                ot[ot.Count - 1].Type = op2.Type;
            }
            else
            {
                ot.Add(opc, rtype, op1, op2);
            }
            ot.Add(OpCodes.ENDEX, VarType.NONE);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <dot-assign-stmt> ::= <identifier> ( '.I' | '.D' | '.S' ) '=' <expression> ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void dot_assign_stmt(string label)
        {
            OpCodes op = OpCodes.NOP;

            switch (la.Peek)
            {
                case Tokens.DOTI: op = OpCodes.MOVI; break;
                case Tokens.DOTS: op = OpCodes.MOVS; break;
                case Tokens.DOTD: op = OpCodes.MOVD; break;
            }
            la.Next();

            la.Next(Tokens.ASSIGN);

            ValType op1 = st.Find(label, func_name);
            ValType op2 = expression(true);
            ot.Add(op, VarType.NONE, op1, op2);

            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <global-array-stmt> ::= <global-array> '=' <expression> ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void global_array_stmt()
        {
            OpCodes op = OpCodes.NOP;
            switch (la.Peek)
            {
                case Tokens.GBYTE: op = OpCodes.WR_BYTE; break;
                case Tokens.GWORD: op = OpCodes.WR_WORD; break;
                case Tokens.GLONG: op = OpCodes.WR_LONG; break;
            }

            la.Next(); // GBYTE | GWORD | GLONG
            la.Next(Tokens.BRACKET_LEFT);

            ValType index = expression(true);
            VarType type = GetType(index);
            if (type != VarType.INT && type != VarType.UINT && type != VarType.AINT)
                throw new Exception("Index must be an integer type");

            la.Next(Tokens.BRACKET_RIGHT);

            if (la.Peek != Tokens.ASSIGN || la.AssignOp != Tokens.ASSIGN)
                throw new Exception("Global write can only be assigned");
            la.Next(Tokens.ASSIGN);

            ValType value = expression(false);
            ot.Add(op, value.Type, value, index);
            ot.Add(OpCodes.ENDEX, VarType.NONE);

            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <global-array> := ( GBYTE | GWORD | GLONG ) '[' <expression> ']' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType global_array()
        {
            VarType type1 = VarType.ANY;
            OpCodes op = OpCodes.NOP;

            switch (la.Peek)
            {
                case Tokens.GBYTE: op = OpCodes.RD_BYTE; break;
                case Tokens.GWORD: op = OpCodes.RD_WORD; break;
                case Tokens.GLONG: op = OpCodes.RD_LONG; break;
            }

            la.Next(); // GBYTE | GWORD | GLONG
            la.Next(Tokens.BRACKET_LEFT);

            ValType index = expression(false);
            VarType type2 = GetType(index);
            if (type2 != VarType.UINT && type2 != VarType.INT && type2 != VarType.AINT)
                throw new Exception("Array index must be an integer type");

            ValType rval = ot.Add(op, type1, st.GetTemp(func_name, type1), index);
            ot.Add(OpCodes.ENDEX, VarType.NONE);
            la.Next(Tokens.BRACKET_RIGHT);

            return rval;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <wait-count-stmt> ::= WAITCNT '(' <identifier> ',' <expression> ')' ';'
        //                     | WAITCNT '(' <expression> ')' ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void wait_count_stmt()
        {
            ValType op1, op2;
            Tokens tk1;

            la.Next(Tokens.WAITCNT);
            la.Next(Tokens.PAREN_LEFT);

            tk1 = la.Peek;
            op1 = expression(true);

            if (la.Peek == Tokens.COMMA)
            {
                la.Next(Tokens.COMMA);

                if (tk1 != Tokens.LABEL)
                    throw new Exception("Expected an identifier");

                op2 = expression(true);
            }
            else
            {
                op2 = st.Find("CONST_0", "");
            }

            ot.Add(OpCodes.WAITCNT, VarType.NONE, op1, op2);

            la.Next(Tokens.PAREN_RIGHT);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <wait-stmt> ::= ( WAITPEQ | WAITPNE | WAITVID ) '(' <expression> ',' <expression> ')' ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void wait_stmt()
        {
            OpCodes op = OpCodes.NOP;
            switch(la.Peek)
            {
                case Tokens.WAITPEQ: op = OpCodes.WAITPEQ; break;
                case Tokens.WAITPNE: op = OpCodes.WAITPNE; break;
                case Tokens.WAITVID: op = OpCodes.WAITVID; break;
            }

            la.Next(); // WAITxxx
            la.Next(Tokens.PAREN_LEFT);

            ValType op1 = expression(true);

            la.Next(Tokens.COMMA);

            ValType op2 = expression(true);

            ot.Add(op, VarType.NONE, op1, op2);

            la.Next(Tokens.PAREN_RIGHT);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <lock-stmt> ::= ( LOCKNEW | LOCKSET | LOCKCLR | LOCKRET ) '(' <IDENTIFIER> ')' ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void lock_stmt()
        {
            OpCodes op = OpCodes.NOP;
            OpCodeEntry.ComparisonT comparison = OpCodeEntry.ComparisonT.NONE;

            switch (la.Peek)
            {
                case Tokens.LOCKNEW: op = OpCodes.LOCKNEW; break;
                case Tokens.LOCKRET: op = OpCodes.LOCKRET; break;
                case Tokens.LOCKSET: op = OpCodes.LOCKSET; comparison = OpCodeEntry.ComparisonT.LS; break;
                case Tokens.LOCKCLR: op = OpCodes.LOCKCLR; comparison = OpCodeEntry.ComparisonT.LC; break;
            }

            la.Next(); // LOCKxxx
            la.Next(Tokens.PAREN_LEFT);

            if (la.Peek != Tokens.LABEL)
                throw new Exception("Expected an identifier");

            ValType op1 = st.Find(la.Label, func_name);
            la.Next(Tokens.LABEL);

            ot.Add(op, VarType.NONE, op1, new ValType(), comparison);

            la.Next(Tokens.PAREN_RIGHT);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <built-in-op-stmt> ::= ( CMPSUB | REV ) '(' <identifier> ',' <expression> ')' ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void biop_stmt()
        {
            OpCodes op;
            switch (la.BIF_Type)
            {
                case BIFType.CMPSUB: op = OpCodes.CMPSUB; break;
                case BIFType.REV: op = OpCodes.REV; break;
                default:
                    throw new Exception("Invalid use of built in function");
            }

            la.Next(Tokens.BIOP);
            la.Next(Tokens.PAREN_LEFT);

            if (la.Peek != Tokens.LABEL)
                throw new Exception("Expected an identifier");
            ValType op1 = st.Find(la.Label, func_name);
            la.Next(Tokens.LABEL);
            la.Next(Tokens.COMMA);

            ValType op2 = expression(true);

            ot.Add(op, VarType.NONE, op1, op2);

            la.Next(Tokens.PAREN_RIGHT);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <nop-stmt> ::= NOP ';' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void nop_stmt()
        {
            ot.Add(OpCodes.NOP, VarType.NONE);
            la.Next(Tokens.NOP);
            la.Next(Tokens.SEMICOLON);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <logical-or-expression> ::= <logical-and-expression>
        //                           | <logical-or-expression '||' <logical-and-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void logical_or_expression(ref string true_label, ref string false_label)
        {
            ValType op1, op2;
                
            op1 = logical_and_expression(false_label);

            while (true)
            {
                if (la.Peek != Tokens.LOGICAL_OR) return;
                la.Next(Tokens.LOGICAL_OR);

                ot.Add(OpCodes.JMP_T, VarType.NONE, true_label);

                ot.Add(OpCodes.LABEL, VarType.NONE, false_label);
                false_label = GetLabel();

                op2 = logical_and_expression(false_label);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <logical-and-expression> ::= <relational-expression>
        //                            | <logical-and-expression> '&&' <relational-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType logical_and_expression(string false_label)
        {
            ValType op1, op2;

            op1 = relational_expression();

            while (true)
            {
                if (la.Peek != Tokens.LOGICAL_AND) return op1;
                la.Next(Tokens.LOGICAL_AND);

                ot.Add(OpCodes.JMP_F, VarType.NONE, false_label);
                op2 = relational_expression();
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <relational-expression> ::= <expression> '==' <expression>
        //                           | <expression> '!=' <expression>
        //                           | <expression> '<' <expression>
        //                           | <expression> '>' <expression>
        //                           | <expression> '<=' <expression>
        //                           | <expression> '>=' <expression>
        //                           | <lock-expression>
        //                           | <parity-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType relational_expression()
        {
            ValType op1, op2;
            OpCodes op;
            OpCodeEntry.ComparisonT comparison;
            Tokens tk;

            if (la.Peek == Tokens.LOCKSET || la.Peek == Tokens.LOCKCLR)
                return lock_expression();

            if (la.Peek == Tokens.PARITY)
                return parity_expression();

            op1 = expression(true);

            tk = la.Peek;
            if (tk == Tokens.PAREN_RIGHT || tk == Tokens.LOGICAL_AND || tk == Tokens.LOGICAL_OR)
            {
                op2 = st.Find("CONST_fls", "");
                op = OpCodes.CMP_NE;
                comparison = OpCodeEntry.ComparisonT.NE;
            }
            else
            {
                switch (tk)
                {
                    case Tokens.CMP_EQ: op = OpCodes.CMP_EQ; comparison = OpCodeEntry.ComparisonT.EQ; break;
                    case Tokens.CMP_NE: op = OpCodes.CMP_NE; comparison = OpCodeEntry.ComparisonT.NE; break;
                    case Tokens.CMP_GT: op = OpCodes.CMP_GT; comparison = OpCodeEntry.ComparisonT.GT; break;
                    case Tokens.CMP_LT: op = OpCodes.CMP_LT; comparison = OpCodeEntry.ComparisonT.LT; break;
                    case Tokens.CMP_GE: op = OpCodes.CMP_GE; comparison = OpCodeEntry.ComparisonT.GE; break;
                    case Tokens.CMP_LE: op = OpCodes.CMP_LE; comparison = OpCodeEntry.ComparisonT.LE; break;
                    default:
                        throw new Exception("Expected a relational operation symbol");
                }
                la.Next(); // == ...
                op2 = expression(true);
                CheckTypes(op1, op2);
            }

            return ot.Add(op, CheckTypes(op1, op2), op1, op2, comparison);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <parity-expression> ::= <parity-function> ( '==' | '!=' } ( ODD | EVEN ) ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType parity_expression()
        {
            ValType op1, op2;
            bool state = true;
            OpCodes opc = OpCodes.BW_TEST;
            OpCodeEntry.ComparisonT comparison;
            Symbol s = st[0];

            la.Next(Tokens.PARITY);
            la.Next(Tokens.PAREN_LEFT);

            op1 = expression(false);

            la.Next(Tokens.COMMA);

            op2 = expression(false);

            la.Next(Tokens.PAREN_RIGHT);

            switch (la.Peek)
            {
                case Tokens.CMP_EQ: break;
                case Tokens.CMP_NE: state = !state; break;
                default:
                    throw new Exception("Expexted equal or not equal comparison");
            }
            la.Next(); // '==' | '!='

            if (la.Peek != Tokens.LABEL)
                throw new Exception("Expected ODD or EVEN");

            switch (la.Label)
            {
                case "CONST_even": break;
                case "CONST_odd": state = !state; break;
                default:
                    throw new Exception("Expexted ODD or EVEN");
            }
            la.Next(); // ODD | EVEN

            if (state)
                comparison = OpCodeEntry.ComparisonT.LC;
            else
                comparison = OpCodeEntry.ComparisonT.LS;

            return ot.Add(opc, VarType.NONE, op1, op2, comparison);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <lock-expression> ::= ( LOCKSET | LOCKCLR ) '(' <identifier> ')' ( '==' | '!=' } ( TRUE | FALSE ) ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType lock_expression()
        {
            ValType op1;
            bool state = true;
            OpCodes opc;
            OpCodeEntry.ComparisonT comparison;

            if (la.Peek == Tokens.LOCKSET)
            {
                la.Next(Tokens.LOCKSET);
                opc = OpCodes.LOCKSET;
            }
            else
            {
                la.Next(Tokens.LOCKCLR);
                opc = OpCodes.LOCKCLR;
            }

            la.Next(Tokens.PAREN_LEFT);

            if (la.Peek != Tokens.LABEL)
                throw new Exception("Expected an identifier");
            op1 = st.Find(la.Label, func_name);
            la.Next(Tokens.LABEL);

            la.Next(Tokens.PAREN_RIGHT);

            switch (la.Peek)
            {
                case Tokens.CMP_EQ: break;
                case Tokens.CMP_NE: state = !state; break;
                default:
                    throw new Exception("Expexted equal or not equal comparison");
            }
            la.Next(); // '==' | '!='

            if (la.Peek != Tokens.LABEL)
                throw new Exception("Expected TRUE or FALSE");

            switch (la.Label)
            {
                case "CONST_tru": break;
                case "CONST_fls": state = !state; break;
                default:
                    throw new Exception("Expexted TRUE or FALSE");
            }
            la.Next(); // TRUE | FALSE

            if (state) comparison = OpCodeEntry.ComparisonT.LS;
            else comparison = OpCodeEntry.ComparisonT.LC;

            return ot.Add(opc, VarType.NONE, op1, new ValType(), comparison);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <expression> ::= <inclusive-or-expression>
        //                | <expression> '#>' <inclusive-or-expression>
        //                | <expression> '<#' <inclusive-or-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType expression(bool endex)
        {
            ValType op1, op2;
            OpCodes opc;

            op1 = inclusive_or_expression();

            if (la.Peek == Tokens.LIMIT_MAX || la.Peek == Tokens.LIMIT_MIN)
                CheckForTemp(ref op1);

            while (true)
            {
                switch (la.Peek)
                {
                    case Tokens.LIMIT_MAX: opc = OpCodes.LIMIT_MAX; break; 
                    case Tokens.LIMIT_MIN: opc = OpCodes.LIMIT_MIN; break;
                    default: 
                        if (endex) ot.Add(OpCodes.ENDEX, VarType.NONE);
                        return op1;
                }
                la.Next(); // '#>' | /<#'

                op2 = inclusive_or_expression();
                op1 = ot.Add(opc, CheckTypes(op1, op2), op1, op2);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <inclusive-or-expression> ::= <exclusive-or-expression>
        //                             | <inclusive-or-expression> '|' <exclusive-or-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType inclusive_or_expression()
        {
            ValType op1, op2;

            op1 = exclusive_or_expression();
            if (la.Peek == Tokens.BITWISE_OR)
                CheckForTemp(ref op1);

            while (la.Peek == Tokens.BITWISE_OR)
            {
                la.Next(Tokens.BITWISE_OR);
                op2 = exclusive_or_expression();
                op1 = ot.Add(OpCodes.BW_OR, CheckTypes(op1, op2), op1, op2);
            }

            return op1;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <exclusive-or-expression> ::= <and-expression>
        //                             | <exclusive-or-expression> '^' <and-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType exclusive_or_expression()
        {
            ValType op1, op2;

            op1 = and_expression();
            if (la.Peek == Tokens.BITWISE_XOR)
                CheckForTemp(ref op1);

            while (la.Peek == Tokens.BITWISE_XOR)
            {
                la.Next(Tokens.BITWISE_XOR);
                op2 = and_expression();
                op1 = ot.Add(OpCodes.BW_XOR, CheckTypes(op1, op2), op1, op2);
            }

            return op1;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <and-expression> ::= <shift-expression>
        //                    | <and-expression> '&' <shift-expression>
        //                    | <and-expression> '&~' <shift-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType and_expression()
        {
            ValType op1, op2;
            OpCodes op;

            op1 = shift_expression();
            if (la.Peek == Tokens.BITWISE_AND || la.Peek == Tokens.BITWISE_AND_NOT)
                CheckForTemp(ref op1);

            while (true)
            {
                switch (la.Peek)
                {
                    case Tokens.BITWISE_AND: op = OpCodes.BW_AND; break;
                    case Tokens.BITWISE_AND_NOT: op = OpCodes.BW_AND_NOT; break;

                    default:
                        return op1;
                }

                la.Next(); // &, &~
                op2 = shift_expression();
                op1 = ot.Add(op, CheckTypes(op1, op2), op1, op2);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <shift-expression> ::= <additive-expression>
        //                      | <shift-expression> '<<' <additive-expression>
        //                      | <shift-expression> '>>' <additive-expression>
        //                      | <shift-expression> '->' <additive-expression>
        //                      | <shift-expression> '<-' <additive-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType shift_expression()
        {
            ValType op1, op2;
            OpCodes op;
            Tokens tk;

            op1 = additive_expression();

            tk = la.Peek;
            if (tk == Tokens.SHIFT_LEFT ||
                tk == Tokens.SHIFT_RIGHT ||
                tk == Tokens.ROTATE_LEFT ||
                tk == Tokens.ROTATE_RIGHT)
                CheckForTemp(ref op1);

            while (true)
            {
                switch (la.Peek)
                {
                    case Tokens.SHIFT_LEFT: op = OpCodes.SHIFT_LEFT; break;
                    case Tokens.SHIFT_RIGHT: op = OpCodes.SHIFT_RIGHT; break;

                    case Tokens.ROTATE_RIGHT: op = OpCodes.ROTATE_RIGHT; break;
                    case Tokens.ROTATE_LEFT: op = OpCodes.ROTATE_LEFT; break;
                    
                    default:
                        return op1;
                }

                la.Next();
                op2 = additive_expression();
                if (op2.Type != VarType.AINT && op2.Type != VarType.UINT && op2.Type != VarType.INT)
                    throw new Exception("Shift count must be an integer type");
                op1 = ot.Add(op, GetType(op1), op1, op2);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <additive-expression> ::= <multiplicative-expression>
        //                         | <additive-expression> '+' <multiplicative-expression>
        //                         | <additive-expression> '-' <multiplicative-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType additive_expression()
        {
            ValType op1, op2;
            OpCodes op;

            op1 = multiplicative_expression();
            if (la.Peek == Tokens.PLUS || la.Peek == Tokens.MINUS)
                CheckForTemp(ref op1);

            while (true)
            {
                switch (la.Peek)
                {
                    case Tokens.PLUS: op = OpCodes.ADD; break;
                    case Tokens.MINUS: op = OpCodes.SUB; break;

                    default:
                        return op1;
                }

                la.Next();
                op2 = multiplicative_expression();
                op1 = ot.Add(op, CheckTypes(op1, op2), op1, op2);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <multiplicative-expression> ::= <unary-expression>
        //                               | <multiplicative-expression> '*' <unary-expression>
        //                               | <multiplicative-expression> '/' <unary-expression>
        //                               | <multiplicative-expression> '%' <unary-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType multiplicative_expression()
        {
            OpCodes op = OpCodes.NOP;
            Tokens tk;
            string rslt = "";

            ValType op1 = unary_expression();
            VarType type1 = GetType(op1);

            while (true)
            {
                tk = la.Peek;
                if (tk != Tokens.MULTIPLY && tk != Tokens.MODULO && tk != Tokens.DIVIDE) return op1;

                la.Next(); // '*' '/' '%'

                ValType op2 = unary_expression();
                VarType type2 = GetType(op2);

                switch (tk)
                {
                    case Tokens.MULTIPLY: 
                        rslt = "math_r1";
                        if (type1 == VarType.REAL && type2 == VarType.REAL)
                            op = OpCodes.RMUL;
                        else
                            op = OpCodes.IMUL; 
                        break;
                    case Tokens.DIVIDE:
                        if (type1 != VarType.REAL && type2 == VarType.REAL)
                            throw new Exception("Int cannot be divided by real data type");
                        rslt = "math_r1";
                        if (type1 == VarType.REAL && type2 == VarType.REAL)
                            op = OpCodes.RDIV;
                        else
                            op = OpCodes.IDIV; 
                        break;
                    case Tokens.MODULO: 
                        if (type1 == VarType.REAL || type2 == VarType.REAL)
                            throw new Exception("Modulo is not valid for real data type");
                        op = OpCodes.IMOD; 
                        rslt = "math_r2"; 
                        break;
                }

                if (type1 == VarType.REAL) type2 = VarType.REAL;

                ot.Add(OpCodes.MOV, op1.Type, st.Find("math_p1", ""), op1);
                ot.Add(OpCodes.MOV, op2.Type, st.Find("math_p2", ""), op2);
                ValType result_var = st.Find(rslt, "");
                result_var.NewType(type2);
                ot.Add(op, type2, result_var, new ValType());
                op1 = ot.Add(OpCodes.MOV, type2, st.GetTemp(func_name, type2), result_var);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <unary-expression> ::= <primary-expression>
        //                      | ! <primary-expression>
        //                      | ~ <primary-expression>
        //                      | - <primary-expression>
        //                      | + <primary-expression> ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType unary_expression()
        {
            ValType op1, op2;

            switch (la.Peek)
            {
                case Tokens.LOGICAL_NOT:
                    la.Next();
                    op1 = primary_expression();
                    return ot.Add(OpCodes.LG_NOT, op1.Type, op1, new ValType());

                case Tokens.BITWISE_NOT:
                    la.Next();
                    op1 = primary_expression();
                    ValType tmp = ot.Add(OpCodes.MOV, op1.Type, st.GetTemp(func_name, op1.Type), op1);
                    return ot.Add(OpCodes.BW_NOT, op1.Type, tmp, op1);

                case Tokens.MINUS:
                    la.Next();
                    op2 = primary_expression();
                    return ot.Add(OpCodes.NEG, op2.Type, st.GetTemp(func_name, op2.Type), op2);

                case Tokens.PLUS:
                    la.Next();
                    return primary_expression();

                default:
                    return primary_expression();
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <primary-expression> ::= <identifier>
        //                        | <function-call>
        //                        | <global-array>
        //                        | <built-in-function>
        //                        | <built-in-operation>
        //                        | <integer-constant>
        //                        | <real-constant>
        //                        | '(' <expression> ')' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType primary_expression()
        {
            ValType rval;
            string label;

            switch (la.Peek)
            {
                case Tokens.LABEL:
                    label = la.Label;
                    la.Next(Tokens.LABEL);
                    if (la.Peek == Tokens.PAREN_LEFT)
                        rval = function_call(label);
                    else
                        rval = st.Find(label, func_name);
                    break;

                case Tokens.CONSTANT:
                    rval = AddConstant(VarType.AINT, la.Value, la.Label);
                    la.Next();
                    break;

                case Tokens.REAL_CONSTANT:
                    rval = AddConstant(VarType.REAL, la.Value, la.Label);
                    la.Next();
                    break;

                case Tokens.PAREN_LEFT:
                    la.Next(Tokens.PAREN_LEFT);
                    rval = expression(false);
                    la.Next(Tokens.PAREN_RIGHT);
                    break;

                case Tokens.GBYTE:
                case Tokens.GWORD:
                case Tokens.GLONG:
                    rval = global_array();
                    break;

                case Tokens.BIF:
                    rval = built_in_function();
                    break;

                case Tokens.BIOP:
                    rval = built_in_operation();
                    break;

                case Tokens.PARITY:
                    rval = parity_function();
                    break;

                default:
                    throw new Exception("Invalid expression");
            }

            return rval;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        ValType AddConstant(VarType type, int value, string label)
        {
           int offset = st.Contains(label, "CONST");
            if (offset < 0)
                offset = st.Add(label, type, value);
            return new ValType(offset, type);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <built-in-function> ::= ( SIN | COS | TAN | ASIN | ACOS | LOG | EXP | SQRT | ISQRT ) '(' <expression> ')' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType built_in_function()
        {
            OpCodes op = OpCodes.NOP;
            VarType vtype = VarType.REAL;
            VarType rtype = VarType.REAL;

            switch (la.BIF_Type)
            {
                case BIFType.SIN: op = OpCodes.SIN; break;
                case BIFType.COS: op = OpCodes.COS; break;
                case BIFType.TAN: op = OpCodes.TAN; break;
                case BIFType.ASIN: op = OpCodes.ASIN; break;
                case BIFType.ACOS: op = OpCodes.ACOS; break;
                case BIFType.LOG: op = OpCodes.LOG; vtype = VarType.ANY; rtype = VarType.AINT; break;
                case BIFType.EXP: op = OpCodes.EXP; vtype = VarType.AINT; rtype = VarType.ANY; break;
                case BIFType.SQRT: op = OpCodes.SQRT; break;
                case BIFType.ISQRT: op = OpCodes.ISQRT; vtype = VarType.AINT; rtype = VarType.AINT; break;
            }
            la.Next(Tokens.BIF);
            la.Next(Tokens.PAREN_LEFT);

            ValType op2 = expression(false);
            CheckTypes(op2, new ValType(0, vtype));

            ot.Add(OpCodes.MOV, vtype, st.Find("math_p1", ""), op2);

            ValType rparm = st.Find("math_r1", "");
            rparm.NewType(rtype);
            ot.Add(op, rtype, rparm, op2);
            ValType rval = ot.Add(OpCodes.MOV, rtype, st.GetTemp(func_name, rtype), rparm);

            la.Next(Tokens.PAREN_RIGHT);

            return rval;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <parity-function> ::= PARITY '(' <expression> ',' <expression> ')' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType parity_function()
        {
            la.Next(Tokens.PARITY);
            la.Next(Tokens.PAREN_LEFT);

            ValType op1 = expression(false);

            la.Next(Tokens.COMMA);

            ValType op2 = expression(false);

            la.Next(Tokens.PAREN_RIGHT);

            ValType temp = st.GetTemp(func_name, VarType.UINT);
            ot.Add(OpCodes.BW_TEST, VarType.UINT, op1, op2, OpCodeEntry.ComparisonT.LS);
            return ot.Add(OpCodes.SUBX, VarType.UINT, temp, temp);
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        // <built-in-operation> ::= ( ABS | ABSN | ITOR | RTOI | CEIL | FLOOR | TRUNC | ROUND | FRACT ) '(' <expression> ')' ;
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private ValType built_in_operation()
        {
            ValType tmp, op1, op2, rval;
            OpCodes op = OpCodes.NOP;
            VarType rtype = VarType.NONE, vtype = VarType.NONE;
            BIFType bif = la.BIF_Type;

            la.Next(Tokens.BIOP);
            la.Next(Tokens.PAREN_LEFT);

            op2 = expression(false);

            switch (bif)
            {
                case BIFType.ABS: op = OpCodes.ABS; rtype = VarType.ANY; vtype = VarType.ANY; break;
                case BIFType.ABSN: op = OpCodes.ABSN; rtype = VarType.ANY; vtype = VarType.ANY; break;
                case BIFType.ITOR: op = OpCodes.ITOR; rtype = VarType.REAL; vtype = VarType.AINT; break;
                case BIFType.RTOI: op = OpCodes.RTOI; rtype = VarType.AINT; vtype = VarType.REAL; break;
                case BIFType.CEIL: op = OpCodes.CEIL; rtype = VarType.REAL; vtype = VarType.REAL; break;
                case BIFType.FLOOR: op = OpCodes.FLOOR; rtype = VarType.REAL; vtype = VarType.REAL; break;
                case BIFType.TRUNC: op = OpCodes.TRUNC; rtype = VarType.REAL; vtype = VarType.REAL; break;
                case BIFType.ROUND: op = OpCodes.ROUND; rtype = VarType.REAL; vtype = VarType.REAL; break;
                case BIFType.FRACT: op = OpCodes.FRACT; rtype = VarType.REAL; vtype = VarType.REAL; break;
            }
            CheckTypes(op2, new ValType(0, vtype));

            op1 = st.GetTemp(func_name, rtype);

            if (op == OpCodes.ITOR || op == OpCodes.RTOI ||
                op == OpCodes.CEIL || op == OpCodes.FLOOR ||
                op == OpCodes.TRUNC || op == OpCodes.ROUND ||
                op == OpCodes.FRACT)
                tmp = ot.Add(OpCodes.MOV, vtype, op1, op2);
            else
                tmp = op1;

            rval = ot.Add(op, rtype, tmp, op2);

            la.Next(Tokens.PAREN_RIGHT);

            return rval;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private void CheckForTemp(ref ValType op)
        {
            if (op.Type != VarType.OPCODE)
            {
                op = ot.Add(OpCodes.MOV, op.Type, st.GetTemp(func_name, op.Type), op);
            }
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        //-----------------------------------------------------------------------------------------------------------------------------------------
        private string GetLabel()
        {
            string s = ":L" + labelid.ToString();
            labelid++;
            return s;
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private VarType CheckTypes(ValType op1, ValType op2)
        {
            while (op1.Type == VarType.OPCODE)
                op1 = ot[op1.Value].Dst;

            while (op2.Type == VarType.OPCODE)
                op2 = ot[op2.Value].Dst;

            switch (op1.Type)
            {
                case VarType.AINT:
                    if (op2.Type == VarType.UINT || op2.Type == VarType.INT || op2.Type == VarType.AINT || op2.Type == VarType.ANY) return op2.Type;
                    break;
                case VarType.INT:
                    if (op2.Type == VarType.INT || op2.Type == VarType.AINT || op2.Type == VarType.ANY) return op1.Type;
                    break;
                case VarType.UINT:
                    if (op2.Type == VarType.UINT || op2.Type == VarType.AINT || op2.Type == VarType.ANY) return op1.Type;
                    break;
                case VarType.REAL:
                    if (op2.Type == VarType.REAL || op2.Type == VarType.ANY) return op1.Type;
                    break;
                case VarType.ANY:
                    if (op2.Type == VarType.UINT || op2.Type == VarType.INT || op2.Type == VarType.AINT || op2.Type == VarType.REAL) return op2.Type;
                    break;
            }
            throw new Exception("Incompatible data types");
        }

        //-----------------------------------------------------------------------------------------------------------------------------------------
        private VarType GetType(ValType op1)
        {
            while (op1.Type == VarType.OPCODE)
                op1 = ot[op1.Value].Dst;

            return op1.Type;
        }
    }
}
