﻿//-----------------------------------------------------------------------------------------
// File:  Program.cs
// Title: Main Program
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

using ASMProp;

namespace PCC
{
    class Program
    {
        static void Main(string[] args)
        {
            string src_file = null;
            string asm_file;
            string tmp_file;
            bool debug_mode = false;
            bool skip_compiler = false;
            bool skip_optimizer = false;
            bool skip_assembler = false;

            LexAn la = new LexAn();
            FuncTable ft = new FuncTable();
            SymbolTable st = new SymbolTable();
            OpcodeTable ot = new OpcodeTable(la);
            Parser parser = new Parser(ft, ot, st, la);
            Optimizer op = new Optimizer(ot, st);
            CodeGen backend = new CodeGen(ot, st, ft);
            Assembler assembler = new Assembler();

            for (int idx = 0; idx < args.Length; idx++)
            {
                switch (args[idx].ToLower())
                {
                    //case "-d": debug_mode = true; break;
                    case "-l": assembler.ListFile = true; break;
                    case "-b": assembler.BinFile = true; break;
                    case "-sc": skip_compiler = true; break;
                    case "-so": skip_optimizer = true; break;
                    case "-sa": skip_assembler = true; break;

                    default: src_file = args[idx]; break;
                }
            }

            if (src_file == null)
            {
                Console.WriteLine("Usage: propc [options] <file_name>");
                Console.WriteLine("\t-sc\tSkip Compiler");
                Console.WriteLine("\t-so\tSkip Optimizer");
                Console.WriteLine("\t-sa\tSkip Assembler");
                Console.WriteLine("\t-b\tOutput Binary File, instead of Hex file");
                Console.WriteLine("\t-l\tOutput Assembler List File");
            }
            else if (System.IO.File.Exists(src_file) == false)
            {
                Console.WriteLine("ERROR: Can't find file " + src_file);
            }
            else
            {
                try
                {
                    asm_file = src_file;
                    if (skip_compiler == false)
                    {
                        tmp_file = Path.ChangeExtension(src_file, "tmp");
                        la.Scan(src_file, tmp_file);

                        Console.WriteLine("Compiling " + src_file);
                        parser.Parse(src_file);
                        if (parser.ErrorMsg != "")
                            throw new Exception(parser.ErrorMsg);

                        if (skip_optimizer == false) op.Run();

                        asm_file = Path.ChangeExtension(src_file, "a");
                        backend.Run(tmp_file, asm_file);
                        if (backend.ErrorMsg != "")
                            throw new Exception(backend.ErrorMsg);

                        Console.WriteLine(backend.Stats);
                    }

                    if (skip_assembler == false)
                    {
                        Console.WriteLine("Assembling " + asm_file);
                        assembler.Assemble(asm_file);
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine(ex.Message);
                }
            }

            if (debug_mode)
            {
                using (StreamWriter sw = new StreamWriter("debug.txt"))
                {
                    for (int i = 0; i < ot.OpTable.Count; i++)
                        //if (ot.OpTable[i].OpCode != OpCodes.NULL && ot.OpTable[i].OpCode != OpCodes.ENDEX)
                            sw.WriteLine(string.Format("{0,3} {1}", i, ot.OpTable[i]));

                    for (int i = 0; i < st.STable.Count; i++)
                        sw.WriteLine(string.Format("{0,3} {1}", i, st.STable[i]));
                }
                Console.ReadLine();
            }
        }
    }
}
