﻿//-----------------------------------------------------------------------------------------
// File:  SymbolTable.cs
// Title: Symbol Table
//
// Copyright Mike Christle 2016
//-----------------------------------------------------------------------------------------
//                               TERMS OF USE: MIT License
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to the following
// conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies
// or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
// PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
// HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
// OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//-----------------------------------------------------------------------------------------
// Date       Version
// ---------- --------
// 2016-06-01    1.0.0 Initial release
//-----------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PCC
{
    public enum SymState { REMOVED, VOID, INITIALIZED, ARRAY, SPIN, CONST };
    public class Symbol
    {
        public VarType Type;
        public string Name;
        public int Value;
        public SymState State;
        public bool Used;
        public Symbol(string name, string scope, VarType type, int value)
        {
            Name = name;
            if (scope != "") Name = scope + '_' + name;
            Type = type;
            Value = value;
            State = SymState.INITIALIZED;
            Used = false;
        }
        public Symbol(string name, VarType type, int value)
        {
            Name = "CONST_" + name;
            Type = type;
            Value = value;
            State = SymState.CONST;
            Used = false;
        }
        public Symbol(string name, string scope, VarType type)
        {
            Name = name;
            if (scope != "") Name = scope + '_' + name;
            Type = type;
            Value = 0;
            State = SymState.VOID;
            Used = false;
        }

        public Symbol(string name, VarType type)
        {
            Name = name;
            Type = type;
            State = SymState.SPIN;
            Used = false;
        }

        public override string ToString()
        {
            return string.Format("{0} {1} {2}", Name, Type, Value);
        }
    }

    class SymbolTable
    {
        public List<Symbol> STable = new List<Symbol>();

        private int TempCount = 0;

        public SymbolTable()
        {
        }

        public int Add(string name, VarType type)
        {
            STable.Add(new Symbol(name, type));
            return STable.Count - 1;
        }

        public int Add(string name, string scope, VarType type)
        {
            STable.Add(new Symbol(name, scope, type));
            return STable.Count - 1;
        }

        public int Add(string name, string scope, VarType type, int value)
        {
            STable.Add(new Symbol(name, scope, type, value));
            return STable.Count - 1;
        }

        public int Add(string name, VarType type, int value)
        {
            STable.Add(new Symbol(name, type, value));
            return STable.Count - 1;
        }

        public ValType Find(string name, string scope)
        {
            string s;

            // Check for function scope variables
            s = scope + '_' + name;
            for (int i = 0; i < STable.Count; i++)
                if (STable[i].Name == s)
                    return new ValType(i, STable[i].Type);

            // Check for global scope variables
            s = name;
            for (int i = 0; i < STable.Count; i++)
                if (STable[i].Name == s)
                    return new ValType(i, STable[i].Type);

            s = "Symbol is not declared: " + s;
            throw new Exception(s);
        }

        public int Contains(string name, string scope)
        {
            string s;
            if (scope == "")
                s = name;
            else
                s = scope + '_' + name;

            for (int i = 0; i < STable.Count; i++)
                if (STable[i].Name == s)
                    return i;

            return -1;
        }

        public ValType GetTemp(string scope, VarType type)
        {
            string name = '_' + TempCount.ToString();
            ++TempCount;

            // Check for function scope variables
            for (int i = 0; i < STable.Count; ++i)
                if (STable[i].Name == scope + '_' + name)
                    return new ValType(i, type);

            STable.Add(new Symbol(name, scope, type));

            return new ValType(STable.Count - 1, type);
        }

        public void ReleaseTemps()
        {
            TempCount = 0;
        }

        public void Remove(int index)
        {
            STable[index].State = SymState.REMOVED;
        }

        public void MarkUsed(int index)
        {
            STable[index].Used = true;
        }

        public void MarkUsed(string name, string scope)
        {
            MarkUsed(Find(name, scope).Value);
        }

        public Symbol this[int index] { get { return STable[index]; } }
        public int Count { get { return STable.Count; } }

        public void Clear()
        {
            STable.Clear();
        }
    }
}
